# This is a copy of the with_metaclass function from 'six' from the
# development branch.  This fixes a bug in six 1.6.1.
#
# Copyright (c) 2010-2014 Benjamin Peterson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import sys
import weakref
from six import itervalues, string_types
import logging
logger = logging.getLogger('pyutilib.component.core')

__all__ = ['Plugin', 'implements', 'Interface', 'CreatePluginFactory',
           'PluginMeta', 'alias', 'ExtensionPoint', 'SingletonPlugin',
           'PluginFactory', 'PluginError', 'PluginGlobals', 'with_metaclass',
           'IPluginLoader', 'IPluginLoadPath', 'IIgnorePluginWhenLoading',
           'IOptionDataProvider', 'PluginEnvironment']

# print "ZZZ - IMPORTING CORE"


def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""

    # This requires a bit of explanation: the basic idea is to make a
    # dummy metaclass for one level of class instantiation that replaces
    # itself with the actual metaclass.  Because of internal type checks
    # we also need to make sure that we downgrade the custom metaclass
    # for one level to something closer to type (that's why __call__ and
    # __init__ comes back from type etc.).
    class metaclass(meta):
        __call__ = type.__call__
        __init__ = type.__init__

        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)

    return metaclass('temporary_class', None, {})

#
# Plugins define within Pyomo
#


#
# Define the default logging behavior for a given namespace, which is to
# ignore the log messages.
#
def logger_factory(namespace):
    log = logging.getLogger('pyutilib.component.core.' + namespace)

    class NullHandler(logging.Handler):

        def emit(self, record):  # pragma:nocover
            """Do not generate logging record"""

    log.addHandler(NullHandler())
    return log


class PluginError(Exception):
    """Exception base class for plugin errors."""

    def __init__(self, value):
        """Constructor, whose argument is the error message"""
        self.value = value

    def __str__(self):
        """Return a string value for this message"""
        return str(self.value)


class PluginEnvironment(object):

    def __init__(self, name=None, bootstrap=False):
        if bootstrap:
            self.env_id = 1
        else:
            PluginGlobals.env_counter += 1
            self.env_id = PluginGlobals.env_counter
            if name is None:
                name = "env%d" % PluginGlobals.env_counter
            if name in PluginGlobals.env:
                raise PluginError("Environment %s is already defined" % name)
        #
        self.loaders = None
        self.loader_paths = None
        #
        self.name = name
        self.log = logger_factory(self.name)
        if __debug__ and self.log.isEnabledFor(logging.DEBUG):
            self.log.debug("Creating PluginEnvironment %r" % self.name)
        # A dictionary of plugin classes
        #   name -> plugin cls
        self.plugin_registry = {}
        # The interfaces that have been defined
        #   name -> interface cls
        self.interfaces = {}
        # A dictionary of singleton plugin class instance ids
        #   plugin cls -> id
        self.singleton_services = {}
        # A set of nonsingleton plugin instances
        self.nonsingleton_plugins = set()

    def cleanup(self, singleton=True):
        # ZZZ
        # return
        if PluginGlobals is None or PluginGlobals.plugin_instances is None:
            return
        if singleton:
            for id_ in itervalues(self.singleton_services):
                if (id_ in PluginGlobals.plugin_instances and
                        PluginGlobals.plugin_instances[id_] is not None):
                    del PluginGlobals.plugin_instances[id_]
            self.singleton_services = {}
        #
        for id_ in self.nonsingleton_plugins:
            del PluginGlobals.plugin_instances[id_]
        self.nonsingleton_plugins = set()

    def plugins(self):
        for id_ in itervalues(self.singleton_services):
            if (id_ in PluginGlobals.plugin_instances and
                    PluginGlobals.plugin_instances[id_] is not None):
                yield PluginGlobals.plugin_instances[id_]
        for id_ in sorted(self.nonsingleton_plugins):
            yield PluginGlobals.plugin_instances[id_]

    def load_services(self, path=None, auto_disable=False, name_re=True):
        """Load services from IPluginLoader extension points"""

        if self.loaders is None:
            self.loaders = ExtensionPoint(IPluginLoader)
            self.loader_paths = ExtensionPoint(IPluginLoadPath)
        #
        # Construct the search path
        #
        search_path = []
        if path is not None:
            if isinstance(path, string_types):
                search_path.append(path)
            elif type(path) is list:
                search_path += path
            else:
                raise PluginError("Unknown type of path argument: " + str(
                    type(path)))
        for item in self.loader_paths:
            search_path += item.get_load_path()
        self.log.info("Loading services to environment "
                      "%s from search path %s" % (self.name, search_path))
        #
        # Compile the enable expression
        #
        if type(auto_disable) is bool:
            if auto_disable:
                disable_p = re.compile("")
            else:
                disable_p = re.compile("^$")
        else:
            disable_p = re.compile(auto_disable)
        #
        # Compile the name expression
        #
        if type(name_re) is bool:
            if name_re:
                name_p = re.compile("")
            else:  # pragma:nocover
                raise PluginError(
                    "It doesn't make sense to specify name_re=False")
        else:
            name_p = re.compile(name_re)
        #
        for loader in self.loaders:
            loader.load(self, search_path, disable_p, name_p)
        # self.clear_cache()

    def Xclear_cache(self):
        """Clear the cache of active services"""
        self._cache = {}


class ExtensionPoint(object):
    """Marker class for extension points in services."""

    def __init__(self, *args):
        """Create the extension point.

        @param interface: the `Interface` subclass that defines the protocol
            for the extension point
        """
        #
        # Construct the interface, passing in this extension
        #
        nargs = len(args)
        if nargs == 0:
            raise PluginError(
                "Must specify interface class used in the ExtensionPoint")
        self.interface = args[0]
        self.__doc__ = 'List of services that implement `%s`' % \
            self.interface.__name__

    def __iter__(self):
        """Return an iterator to a set of services that match the interface of
        this extension point.
        """
        return self.extensions().__iter__()

    def __call__(self, key=None, all=False):
        """Return a set of services that match the interface of this
        extension point.
        """
        if type(key) in (int, int):
            raise PluginError("Access of the n-th extension point is "
                              "disallowed.  This is not well-defined, since "
                              "ExtensionPoints are stored as unordered sets.")
        return self.extensions(all=all, key=key)

    def service(self, key=None, all=False):
        """Return the unique service that matches the interface of this
        extension point.  An exception occurs if no service matches the
        specified key, or if multiple services match.
        """
        ans = ExtensionPoint.__call__(self, key=key, all=all)
        if len(ans) == 1:
            #
            # There is a single service, so return it.
            #
            return ans.pop()
        elif len(ans) == 0:
            return None
        else:
            raise PluginError("The ExtensionPoint does not have a unique "
                              "service!  %d services are defined for interface"
                              " %s.  (key=%s)" %
                              (len(ans), self.interface.__name__, str(key)))

    def __len__(self):
        """Return the number of services that match the interface of this
        extension point.
        """
        return len(self.extensions())

    def extensions(self, all=False, key=None):
        """Return a set of services that match the interface of this
        extension point.  This tacitly filters out disabled extension
        points.

        TODO - Can this support caching?
        How would that relate to the weakref test?
        """
        strkey = str(key)
        ans = set()
        remove = set()
        if self.interface in PluginGlobals.interface_services:
            for id_ in PluginGlobals.interface_services[self.interface]:
                if id_ not in PluginGlobals.plugin_instances:
                    remove.add(id_)
                    continue
                if id_ < 0:
                    plugin = PluginGlobals.plugin_instances[id_]
                else:
                    plugin = PluginGlobals.plugin_instances[id_]()
                if plugin is None:
                    remove.add(id_)
                elif ((all or plugin.enabled()) and
                      (key is None or strkey == plugin.name)):
                    ans.add(plugin)
            # Remove weakrefs that were empty
            # ZZ
            for id_ in remove:
                PluginGlobals.interface_services[self.interface].remove(id_)
        return sorted(ans, key=lambda x: x._id)

    def __repr__(self, simple=False):
        """Return a textual representation of the extension point.

        TODO: use the 'simple' argument
        """
        env_str = ""
        for env_ in itervalues(PluginGlobals.env):
            if self.interface in set(itervalues(env_.interfaces)):
                env_str += " env=%s" % env_.name
        return '<ExtensionPoint %s%s>' % (self.interface.__name__, env_str)


class PluginGlobals(object):
    """Global data for plugins. The main role of this class is to manage
    the stack of PluginEnvironment instances.

    Note: a single ID counter is used for tagging both environment and
    plugins registrations.  This enables the  user to track the relative
    order of construction of these objects.
    """

    def __init__(self):  # pragma:nocover
        """Disable construction."""
        raise PluginError("The PluginGlobals class should not be created.")

    # A dictionary of interface classes mapped to sets of plugin class
    # instance ids
    #   interface cls -> set(ids)
    interface_services = {}

    # A dictionary of plugin instances
    #   id -> weakref(instance)
    plugin_instances = {}

    # Environments
    env = {'pca': PluginEnvironment('pca', bootstrap=True)}
    env_map = {1: 'pca'}
    env_stack = ['pca']
    """A unique id used to name plugin objects"""
    plugin_counter = 0
    """A unique id used to name environment objects"""
    env_counter = 1
    """A list of executables"""
    _executables = []
    """TODO"""
    _default_OptionData = None

    @staticmethod
    def get_env(arg=None):
        """Return the current environment."""
        if arg is None:
            return PluginGlobals.env[PluginGlobals.env_stack[-1]]
        else:
            return PluginGlobals.env.get(arg, None)

    @staticmethod
    def add_env(name=None, validate=False):
        if name is not None and not isinstance(name, string_types):
            if validate and name.name in PluginGlobals.env:
                raise PluginError("Environment %s is already defined" % name)
            # We assume we have a PluginEnvironment object here
            PluginGlobals.env[name.name] = name
            PluginGlobals.env_map[name.env_id] = name.name
            PluginGlobals.env_stack.append(name.name)
            if __debug__ and name.log.isEnabledFor(logging.DEBUG):
                name.log.debug("Pushing environment %r on the "
                               "PluginGlobals stack" % name.name)
            return name
        else:
            env_ = PluginGlobals.env.get(name, None)
            if validate and env_ is not None:
                raise PluginError("Environment %s is already defined" % name)
            if env_ is None:
                env_ = PluginEnvironment(name)
                PluginGlobals.env[env_.name] = env_
            PluginGlobals.env_map[env_.env_id] = env_.name
            PluginGlobals.env_stack.append(env_.name)
            if __debug__ and env_.log.isEnabledFor(logging.DEBUG):
                env_.log.debug("Pushing environment %r on the "
                               "PluginGlobals stack" % env_.name)
            return env_

    @staticmethod
    def pop_env():
        if len(PluginGlobals.env_stack) > 1:
            name = PluginGlobals.env_stack.pop()
            env_ = PluginGlobals.env[name]
            if __debug__ and env_.log.isEnabledFor(logging.DEBUG):
                env_.log.debug("Popping environment %r from the "
                               "PluginGlobals stack" % env_.name)
            return env_
        else:
            return PluginGlobals.env[PluginGlobals.env_stack[0]]

    @staticmethod
    def remove_env(name, cleanup=False, singleton=True):
        tmp = PluginGlobals.env.get(name, None)
        if tmp is None:
            raise PluginError("No environment %s is defined" % name)
        # print "HERE - remove", name, tmp.env_id
        del PluginGlobals.env_map[tmp.env_id]
        del PluginGlobals.env[name]
        if cleanup:
            tmp.cleanup(singleton=singleton)
        PluginGlobals.env_stack = [name_ for name_ in PluginGlobals.env_stack
                                   if name_ in PluginGlobals.env]
        return tmp

    @staticmethod
    def clear():
        # ZZ
        # return
        for env_ in itervalues(PluginGlobals.env):
            env_.cleanup()
        PluginGlobals.interface_services = {}
        PluginGlobals.plugin_instances = {}
        PluginGlobals.env = {'pca': PluginEnvironment('pca', bootstrap=True)}
        PluginGlobals.env_map = {1: 'pca'}
        PluginGlobals.env_stack = ['pca']
        PluginGlobals.plugin_counter = 0
        PluginGlobals.env_counter = 1
        PluginGlobals._executables = []

    @staticmethod
    def clear_global_data(keys=None):
        # ZZ
        # return
        ep = ExtensionPoint(IOptionDataProvider)
        for ep_ in ep:
            ep_.clear(keys=keys)

    @staticmethod
    def services(name=None):
        """A convenience function that returns the services in the
        current environment.

        TODO:  env-specific services?
        """
        ans = set()
        for ids in itervalues(PluginGlobals.interface_services):
            for id_ in ids:
                if id_ not in PluginGlobals.plugin_instances:
                    # TODO: discard the id from the set?
                    continue
                if id_ < 0:
                    ans.add(PluginGlobals.plugin_instances[id_])
                else:
                    ans.add(PluginGlobals.plugin_instances[id_]())
        return ans

    @staticmethod
    def load_services(**kwds):
        """Load services from IPluginLoader extension points"""
        PluginGlobals.get_env().load_services(**kwds)

    @staticmethod
    def pprint(**kwds):
        """A pretty-print function"""

        ans = {}
        s = "#------------------------------"\
            "--------------------------------\n"
        i = 1
        ans['Environment Stack'] = {}
        s += "Environment Stack:\n"
        for env in PluginGlobals.env_stack:
            ans['Environment Stack'][i] = env
            s += "  '" + str(i) + "': " + env + "\n"
            i += 1
        s += "#------------------------------"\
             "--------------------------------\n"
        ans['Interfaces Declared'] = {}
        s += "Interfaces Declared:\n"
        keys = []
        for env_ in itervalues(PluginGlobals.env):
            keys.extend(interface_ for interface_ in env_.interfaces)
        keys.sort()
        for key in keys:
            ans['Interfaces Declared'][key] = None
            s += "  " + key + ":\n"
        s += "#------------------------------"\
             "--------------------------------\n"
        ans['Interfaces Declared by Environment'] = {}
        s += "Interfaces Declared by Environment:\n"
        for env_name in sorted(PluginGlobals.env.keys()):
            env_ = PluginGlobals.env[env_name]
            if len(env_.interfaces) == 0:
                continue
            ans['Interfaces Declared by Environment'][env_.name] = {}
            s += "  " + env_.name + ":\n"
            for interface_ in sorted(env_.interfaces.keys()):
                ans['Interfaces Declared by Environment'][env_.name][
                    interface_] = None
                s += "    " + interface_ + ":\n"
        #
        # Coverage is disabled here because different platforms give different
        # results.
        #
        if kwds.get('plugins', True):  # pragma:nocover
            s += "#---------------------------"\
                 "-----------------------------------\n"
            ans['Plugins by Environment'] = {}
            s += "Plugins by Environment:\n"
            for env_name in sorted(PluginGlobals.env.keys()):
                env_ = PluginGlobals.env[env_name]
                ans['Plugins by Environment'][env_.name] = {}
                s += "  " + env_.name + ":\n"
                flag = True
                for service_ in env_.plugins():
                    flag = False
                    try:
                        service_.name
                    except:
                        service_ = service_()
                    service_active = False
                    for interface in service_.__interfaces__:
                        if interface not in PluginGlobals.interface_services:
                            continue
                        service_active = service_._id in \
                            PluginGlobals.interface_services[interface]
                        break
                    service_s = service_.__repr__(
                        simple=not kwds.get('show_ids', True))
                    ans['Plugins by Environment'][env_.name][service_s] = {}
                    s += "    " + service_s + ":\n"
                    if kwds.get('show_ids', True):
                        ans['Plugins by Environment'][env_.name][service_s][
                            'name'] = service_.name
                        s += "       name:      " + service_.name + "\n"
                    ans['Plugins by Environment'][env_.name][service_s][
                        'id'] = str(service_._id)
                    s += "       id:        " + str(service_._id) + "\n"
                    ans['Plugins by Environment'][env_.name][service_s][
                        'singleton'] = str(service_.__singleton__)
                    s += "       singleton: " + \
                        str(service_.__singleton__) + "\n"
                    ans['Plugins by Environment'][env_.name][service_s][
                        'service'] = str(service_active)
                    s += "       service:   " + str(service_active) + "\n"
                    ans['Plugins by Environment'][env_.name][service_s][
                        'disabled'] = str(not service_.enabled())
                    s += "       disabled:  " + \
                        str(not service_.enabled()) + "\n"
                if flag:
                    s += "       None:\n"
        s += "#------------------------------"\
             "--------------------------------\n"
        ans['Plugins by Interface'] = {}
        s += "Plugins by Interface:\n"
        tmp = {}
        for env_ in itervalues(PluginGlobals.env):
            for interface_ in itervalues(env_.interfaces):
                tmp[interface_] = []
        for env_ in itervalues(PluginGlobals.env):
            for key in env_.plugin_registry:
                for item in env_.plugin_registry[key].__interfaces__:
                    if item in tmp:
                        tmp[item].append(key)
        keys = list(tmp.keys())
        for key in sorted(keys, key=lambda v: v.__name__.upper()):
            ans['Plugins by Interface'][str(key.__name__)] = {}
            if key.__name__ == "":  # pragma:nocover
                s += "  `" + str(key.__name__) + "`:\n"
            else:
                s += "  " + str(key.__name__) + ":\n"
            ttmp = tmp[key]
            ttmp.sort()
            if len(ttmp) == 0:
                s += "    None:\n"
            else:
                for item in ttmp:
                    ans['Plugins by Interface'][str(key.__name__)][item] = None
                    s += "    " + item + ":\n"
        s += "#-------------------------------"\
             "-------------------------------\n"
        ans['Plugins by Python Module'] = {}
        s += "Plugins by Python Module:\n"
        tmp = {}
        for env_ in itervalues(PluginGlobals.env):
            for name_ in env_.plugin_registry:
                tmp.setdefault(env_.plugin_registry[name_].__module__,
                               []).append(name_)
        if '__main__' in tmp:
            # This is a hack to ensure consistency in the tests
            tmp['pyutilib.component.core.tests.test_core'] = tmp['__main__']
            del tmp['__main__']
        keys = list(tmp.keys())
        keys.sort()
        for key in keys:
            ans['Plugins by Python Module'][str(key)] = {}
            if key == "":  # pragma:nocover
                s += "  `" + str(key) + "`:\n"
            else:
                s += "  " + str(key) + ":\n"
            ttmp = tmp[key]
            ttmp.sort()
            for item in ttmp:
                ans['Plugins by Python Module'][str(key)][item] = None
                s += "    " + item + ":\n"
        if kwds.get('json', False):
            import json
            print(
                json.dumps(
                    ans, sort_keys=True, indent=4, separators=(',', ': ')))
        else:
            print(s)

    @staticmethod
    def display(interface=None, verbose=False):
        print("Plugin Instances:", len(PluginGlobals.plugin_instances))
        if interface is not None:
            print("Interface:", interface.name)
            print("Count:",
                  len(PluginGlobals.interface_services.get(interface, [])))
            if verbose:
                for service in interface.services.get(interface, []):
                    print(service)
        else:
            print("Interfaces", len(PluginGlobals.interface_services))
            for interface in PluginGlobals.interface_services:
                print("  Interface:", interface)
                print("  Count:",
                      len(PluginGlobals.interface_services.get(interface, [])))
                if verbose:
                    for service in PluginGlobals.interface_services.get(
                            interface, []):
                        print("     ", PluginGlobals.plugin_instances[service])
        #
        print("")
        for env_ in itervalues(PluginGlobals.env):
            print("Plugin Declarations:", env_.name)
            for interface in sorted(
                    env_.interfaces.keys(), key=lambda v: v.upper()):
                print("Interface:", interface)
                # print "Aliases:"
                # for alias in sorted(interface._factory_cls.keys(),
                #                     key=lambda v: v.upper()):
                #     print "   ",alias,interface._factory_cls[alias]


class InterfaceMeta(type):
    """Meta class that registered the declaration of an interface"""

    def __new__(cls, name, bases, d):
        """Register this interface"""
        new_class = type.__new__(cls, name, bases, d)
        if name != "Interface":
            if name in PluginGlobals.get_env().interfaces:
                raise PluginError("Interface %s has already been defined" %
                                  name)
            PluginGlobals.get_env().interfaces[name] = new_class
        return new_class


class Interface(with_metaclass(InterfaceMeta, object)):
    """Marker base class for extension point interfaces.  This class
    is not intended to be instantiated.  Instead, the declaration
    of subclasses of Interface are recorded, and these
    classes are used to define extension points.
    """
    pass


class IPluginLoader(Interface):
    """An interface for loading plugins."""

    def load(self, env, path, disable_re, name_re):
        """Load plugins found on the specified path.  If disable_re is
        not none, then it is interpreted as a regular expression.  If
        this expression matches the path of a plugin, then that plugin
        is disabled. Otherwise, the plugin is enabled by default.
        """


class IPluginLoadPath(Interface):

    def get_load_path(self):
        """Returns a list of paths that are searched for plugins"""


class IIgnorePluginWhenLoading(Interface):
    """Interface used by Plugin loaders to identify Plugins that should
    be ignored"""

    def ignore(self, name):
        """Returns true if a loader should ignore a plugin during loading"""


class IOptionDataProvider(Interface):
    """An interface that supports the management of common data between
    Options.  This interface is also used to share this data with
    the Configuration class.
    """

    def get_data(self):
        """Returns a dictionary of dictionaries that represents the
        options data."""

    def set(self, section, name, value):
        """Sets the value of a given (section,name) pair"""

    def get(self, section, name):
        """Returns the value of a given (section,name) pair"""

    def clear(self):
        """Clears the data"""


class PluginMeta(type):
    """Meta class for the Plugin class.  This meta class
    takes care of service and extension point registration.  This class
    also instantiates singleton plugins.
    """

    def __new__(cls, name, bases, d):
        """Find all interfaces that need to be registered."""
        #
        # Avoid cycling in the Python logic by hard-coding the behavior
        # for the Plugin and SingletonPlugin classes.
        #
        if name == "Plugin":
            d['__singleton__'] = False
            return type.__new__(cls, name, bases, d)
        if name == "SingletonPlugin":
            d['__singleton__'] = True
            return type.__new__(cls, name, bases, d)
        if name == "ManagedSingletonPlugin":
            #
            # This is a derived class of SingletonPlugin for which
            # we do not need to build an instance
            #
            d['__singleton__'] = True
            return type.__new__(cls, name, bases, d)
        #
        # Check if plugin has already been registered
        #
        if len(d.get('_implements', [])) == 0 and (
                name in PluginGlobals.get_env().plugin_registry):
            return PluginGlobals.get_env().plugin_registry[name]
        #
        # Find all interfaces that this plugin will support
        #
        __interfaces__ = {}
        for interface in d.get('_implements', {}):
            __interfaces__.setdefault(interface,
                                      []).extend(d['_implements'][interface])
        for base in [base for base in bases if hasattr(base, '__interfaces__')]:
            for interface in base.__interfaces__:
                __interfaces__.setdefault(
                    interface, []).extend(base.__interfaces__[interface])
        d['__interfaces__'] = __interfaces__
        #
        # Create a boolean, which indicates whether this is
        # a singleton class.
        #
        if True in [issubclass(x, SingletonPlugin) for x in bases]:
            d['__singleton__'] = True
        else:
            d['__singleton__'] = False
        #
        # Add interfaces to the list of base classes if they are
        # declared inherited.
        #
        flag = False
        bases = list(bases)
        for interface in d.get('_inherited_interfaces', set()):
            if interface not in bases:
                bases.append(interface)
                flag = True
        if flag:
            cls = MergedPluginMeta
        #
        # Create new class
        #
        try:
            new_class = type.__new__(cls, name, tuple(bases), d)
        except:
            raise
        setattr(new_class, '__name__', name)
        #
        for _interface in __interfaces__:
            if getattr(_interface, '_factory_active', None) is None:
                continue
            for _name, _doc, _subclass in getattr(new_class, "_factory_aliases",
                                                  []):
                if _name in _interface._factory_active:
                    if _subclass:
                        continue
                    else:
                        raise PluginError("Alias '%s' has already been "
                                          "defined for interface '%s'" %
                                          (_name, str(_interface)))
                _interface._factory_active[_name] = name
                _interface._factory_doc[_name] = _doc
                _interface._factory_cls[_name] = new_class
        #
        if d['__singleton__']:
            #
            # Here, we create an instance of a singleton class, which
            # registers itself in singleton_services
            #
            PluginGlobals.get_env().singleton_services[new_class] = True
            __instance__ = new_class()
            PluginGlobals.plugin_instances[__instance__._id] = __instance__
            PluginGlobals.get_env().singleton_services[new_class] = \
                __instance__._id
        else:
            __instance__ = None
        #
        # Register this plugin
        #
        PluginGlobals.get_env().plugin_registry[name] = new_class
        return new_class


class MergedPluginMeta(PluginMeta, InterfaceMeta):

    def __new__(cls, name, bases, d):
        return PluginMeta.__new__(cls, name, bases, d)


class Plugin(with_metaclass(PluginMeta, object)):
    """Base class for plugins.  A 'service' is an instance of a Plugin.

    Every Plugin class can declare what extension points it provides, as
    well as what extension points of other Plugins it extends.
    """

    def __del__(self):
        # print "HERE - plugin __del__", self._id, self.name,
        #       self.__class__.__name__
        # ZZZ
        # return
        self.deactivate()
        if (PluginGlobals is not None and
                PluginGlobals.plugin_instances is not None and
                self._id in PluginGlobals.plugin_instances and
                PluginGlobals.plugin_instances[self._id] is not None):
            # print "HERE - plugin __del__", self._id
            # print "interface_services", PluginGlobals.interface_services
            # print "HERE", self.name, self.__class__.__name__
            del PluginGlobals.plugin_instances[self._id]
        if (PluginGlobals is not None and PluginGlobals.env_map is not None and
                self._id_env in PluginGlobals.env_map):
            PluginGlobals.env[PluginGlobals.env_map[
                self._id_env]].nonsingleton_plugins.discard(self._id)

    def __init__(self, **kwargs):
        if "name" in kwargs:
            self.name = kwargs["name"]

    def __new__(cls, *args, **kwargs):
        """Plugin constructor"""
        #
        # If this service is a singleton, then allocate and configure
        # it differently.
        #
        if cls in PluginGlobals.get_env().singleton_services:  # pragma:nocover
            id = PluginGlobals.get_env().singleton_services[cls]
            if id is True:
                self = super(Plugin, cls).__new__(cls)
                PluginGlobals.plugin_counter += 1
                self._id = -PluginGlobals.plugin_counter
                self._id_env = PluginGlobals.get_env().env_id
                self.name = self.__class__.__name__
                self._enable = True
                self.activate()
            else:
                self = PluginGlobals.plugin_instances[id]
            # print "HERE - Plugin singleton:",
            #       self._id, self.name, self._id_env
            return self
        #
        # Else we generate a normal plugin
        #
        self = super(Plugin, cls).__new__(cls)
        PluginGlobals.plugin_counter += 1
        self._id = PluginGlobals.plugin_counter
        self._id_env = PluginGlobals.get_env().env_id
        self.name = "Plugin." + str(self._id)
        PluginGlobals.get_env().nonsingleton_plugins.add(self._id)
        # print "HERE - Normal Plugin:", self._id, self.name,
        #   self.__class__.__name__, self._id_env
        self._enable = True
        PluginGlobals.plugin_instances[self._id] = weakref.ref(self)
        if getattr(cls, '_service', True):
            # self._HERE_ = self._id
            self.activate()
        return self

    def activate(self):
        """Register this plugin with all interfaces that it implements."""
        for interface in self.__interfaces__:
            PluginGlobals.interface_services.setdefault(interface,
                                                        set()).add(self._id)

    def deactivate(self):
        """Unregister this plugin with all interfaces that it implements."""
        # ZZ
        # return
        if PluginGlobals is None or PluginGlobals.interface_services is None:
            # This could happen when python quits
            return
        # for interface in self.__interfaces__:
        # if interface in PluginGlobals.interface_services:
        for interface in PluginGlobals.interface_services:
            # Remove an element if it exists
            PluginGlobals.interface_services[interface].discard(self._id)

    #
    # Support "with" statements. Forgetting to call deactivate
    # on Plugins is a common source of memory leaks
    #
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.deactivate()

    @staticmethod
    def alias(name, doc=None, subclass=False):
        """This function is used to declare aliases that can be used by a
        factory for constructing plugin instances.

        When the subclass option is True, then subsequent calls to
        alias() with this class name are ignored, because they are
        assumed to be due to subclasses of the original class declaration.
        """
        frame = sys._getframe(1)
        locals_ = frame.f_locals
        assert locals_ is not frame.f_globals and '__module__' in locals_, \
            'alias() can only be used in a class definition'
        locals_.setdefault('_factory_aliases', set()).add((name, doc, subclass))

    @staticmethod
    def implements(interface, inherit=None, namespace=None, service=False):
        """Can be used in the class definition of `Plugin` subclasses to
        declare the extension points that are implemented by this
        interface class.
        """
        frame = sys._getframe(1)
        locals_ = frame.f_locals
        #
        # Some sanity checks
        #
        assert namespace is None or isinstance(namespace, str), \
            'second implements() argument must be a string'
        assert locals_ is not frame.f_globals and '__module__' in locals_, \
            'implements() can only be used in a class definition'
        #
        locals_.setdefault('_implements', {}).setdefault(interface,
                                                         []).append(namespace)
        if inherit:
            locals_.setdefault('_inherited_interfaces', set()).add(interface)
        locals_['_service'] = service

    def disable(self):
        """Disable this plugin"""
        self._enable = False

    def enable(self):
        """Enable this plugin"""
        self._enable = True

    def enabled(self):
        """Return value indicating if this plugin is enabled"""
        enable = self._enable
        if hasattr(enable, 'get_value'):
            try:
                enable = enable.get_value()
            except:
                enable = None
        return enable

    def __repr__(self, simple=False):
        """Return a textual representation of the plugin."""
        if simple or self.__class__.__name__ == self.name:
            return '<Plugin %s>' % (self.__class__.__name__)
        else:
            return '<Plugin %s %r>' % (self.__class__.__name__, self.name)


alias = Plugin.alias
implements = Plugin.implements


class SingletonPlugin(Plugin):
    """The base class for singleton plugins.  The PluginMeta class
    instantiates a SingletonPlugin class when it is declared.  Note that
    only one instance of a SingletonPlugin class is created in
    any environment.
    """
    pass


def CreatePluginFactory(_interface):
    if getattr(_interface, '_factory_active', None) is None:
        setattr(_interface, '_factory_active', {})
        setattr(_interface, '_factory_doc', {})
        setattr(_interface, '_factory_cls', {})
        setattr(_interface, '_factory_deactivated', {})

    class PluginFactoryFunctor(object):

        def __call__(self, _name=None, args=[], **kwds):
            if _name is None:
                return self
            _name = str(_name)
            if _name not in _interface._factory_active:
                return None
            return PluginFactory(_interface._factory_cls[_name], args, **kwds)

        def services(self):
            return list(_interface._factory_active.keys())

        def get_class(self, name):
            return _interface._factory_cls[name]

        def doc(self, name):
            tmp = _interface._factory_doc[name]
            if tmp is None:
                return ""
            return tmp

        def deactivate(self, name):
            if name in _interface._factory_active:
                _interface._factory_deactivated[
                    name] = _interface._factory_active[name]
                del _interface._factory_active[name]

        def activate(self, name):
            if name in _interface._factory_deactivated:
                _interface._factory_active[
                    name] = _interface._factory_deactivated[name]
                del _interface._factory_deactivated[name]

    return PluginFactoryFunctor()


def PluginFactory(classname, args=[], **kwds):
    """Construct a Plugin instance, and optionally assign it a name"""

    if isinstance(classname, str):
        try:
            cls = PluginGlobals.get_env(kwds.get('env', None)).plugin_registry[
                classname]
        except KeyError:
            raise PluginError("Unknown class %r" % str(classname))
    else:
        cls = classname
    obj = cls(*args, **kwds)
    if 'name' in kwds:
        obj.name = kwds['name']
    if __debug__ and logger.isEnabledFor(logging.DEBUG):
        if obj is None:
            logger.debug("Failed to create plugin %s" % (classname))
        else:
            logger.debug("Creating plugin %s with name %s" % (classname,
                                                              obj.name))
    return obj



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

import re
import copy
import sys
import os
import os.path
import difflib
import zipfile
import gzip
import filecmp
import math
if sys.version_info >= (3, 0):
    xrange = range
    import io

strict_float_p = re.compile(
    r"(?<![\w+-\.])(?:[+-])?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?\b")
relaxed_float_p = re.compile(
    r"(?:[+-])?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?")
whitespace_p = re.compile(r" +")


def remove_chars_in_list(s, l):
    if len(l) == 0:
        return s

    return "".join(x for x in s if x not in l)


def get_desired_chars_from_file(f, nchars, l=""):
    retBuf = ""
    while nchars > 0:
        buf = f.read(nchars)
        if len(buf) == 0:
            break

        buf = remove_chars_in_list(buf, l)
        nchars -= len(buf)
        retBuf = retBuf + buf

    return retBuf


if sys.version_info[:2] == (3, 2):
    #
    # This fixes a bug in Python 3.2's implementation of GzipFile.
    #
    class MyGzipFile(gzip.GzipFile):

        def read1(self, n):
            return self.read(n)


def open_possibly_compressed_file(filename):
    if not os.path.exists(filename):
        raise IOError("cannot find file `" + filename + "'")
    if sys.version_info[:2] < (2, 6) and zipfile.is_zipfile(filename):
        raise IOError("cannot unpack a ZIP file with Python %s" %
                      '.'.join(map(str, sys.version_info)))
    try:
        is_zipfile = zipfile.is_zipfile(filename)
    except:
        is_zipfile = False
    if is_zipfile:
        zf1 = zipfile.ZipFile(filename, "r")
        if len(zf1.namelist()) != 1:
            raise IOError("cannot compare with a zip file that contains "
                          "multiple files: `" + filename + "'")
        if sys.version_info < (3, 0):
            return zf1.open(zf1.namelist()[0], 'r')
        else:
            return io.TextIOWrapper(
                zf1.open(zf1.namelist()[0], 'r'), encoding='utf-8', newline='')
    elif filename.endswith('.gz'):
        if sys.version_info < (3, 0):
            return gzip.open(filename, "r")
        elif sys.version_info[:2] == (3, 2):
            return io.TextIOWrapper(
                MyGzipFile(filename), encoding='utf-8', newline='')
        else:
            return io.TextIOWrapper(
                gzip.open(filename, 'r'), encoding='utf-8', newline='')
    else:
        return open(filename, "r")


def file_diff(filename1, filename2, lineno=None, context=None):
    INPUT1 = open_possibly_compressed_file(filename1)
    lines1 = INPUT1.readlines()
    for i in range(0, len(lines1)):
        lines1[i] = lines1[i].strip()
    INPUT1.close()

    INPUT2 = open_possibly_compressed_file(filename2)
    lines2 = INPUT2.readlines()
    for i in range(0, len(lines2)):
        lines2[i] = lines2[i].strip()
    INPUT2.close()

    s = ""
    if lineno is None:
        for line in difflib.unified_diff(
                lines2, lines1, fromfile=filename2, tofile=filename1):
            s += line + "\n"
    else:
        if context is None:
            context = 3
        start = lineno - context
        stop = lineno + context
        if start < 0:
            start = 0
        if stop > len(lines1):
            stop = len(lines1)
        if stop > len(lines2):
            stop = len(lines2)
        for line in difflib.unified_diff(
                lines2[start:stop],
                lines1[start:stop],
                fromfile=filename2,
                tofile=filename1):
            s += line + "\n"
    return s


def read_and_filter_line(stream, ignore_chars, filter):
    # If either line is composed entirely of characters to
    # ignore, then get another one.  In this way we can
    # skip blank lines that are in one file but not the other
    lineno = 0
    line = ""
    while not line:
        line = stream.readline()
        lineno += 1
        if line == "":
            return None, lineno
        line_ = remove_chars_in_list(line, ignore_chars)
        if not line_:
            line = False
            continue
        if filter is not None:
            filtered = filter(line)
            if filtered is True:
                line = False  # Ignore this line
            elif filtered is False:
                line = line_
            else:
                line = filtered
        else:
            line = line_
    return line, lineno


def _extract_floats(line, regex):
    ans = []
    while True:
        g = regex.search(line)
        if g is None:
            return ans, line
        ans.append(float(g.group()))
        line = regex.sub(" # ", line, count=1)


def compare_file_with_numeric_values(filename1,
                                     filename2,
                                     ignore=["\n", "\r"],
                                     filter=None,
                                     tolerance=0.0,
                                     strict_numbers=True):
    """
    Do a simple comparison of two files that ignores differences
    in newline types and whitespace.  Numeric values are compared within a specified tolerance.

    The return value is the tuple: (status,lineno).  If status is True,
    then a difference has occured on the specified line number.  If
    the status is False, then lineno is None.

    The goal of this utility is to simply indicate whether there are
    differences in files.  The Python 'difflib' is much more comprehensive
    and consequently more costly to apply.  The shutil.filecmp utility is
    similar, but it does not ignore differences in file newlines.  Also,
    this utility can ignore an arbitrary set of characters.
    """
    if not os.path.exists(filename1):
        raise IOError("compare_file: cannot find file `" + filename1 + "' (in "
                      + os.getcwd() + ")")
    if not os.path.exists(filename2):
        raise IOError("compare_file: cannot find file `" + filename2 + "' (in "
                      + os.getcwd() + ")")

    #if filecmp.cmp(filename1, filename2):
    #    return [False, None, ""]

    if strict_numbers:
        float_p = strict_float_p
    else:
        float_p = relaxed_float_p

    try:
        absolute_tolerance, relative_tolerance = tolerance
    except:
        absolute_tolerance = relative_tolerance = tolerance

    INPUT1 = INPUT2 = None
    try:
        INPUT1 = open_possibly_compressed_file(filename1)
        INPUT2 = open_possibly_compressed_file(filename2)

        lineno = 0
        while True:

            # If either line is composed entirely of characters to
            # ignore, then get another one.  In this way we can
            # skip blank lines that are in one file but not the other

            try:
                line1, delta_lineno = read_and_filter_line(
                    INPUT1, ignore, filter)
            except UnicodeDecodeError:
                err = sys.exc_info()[1]
                raise RuntimeError(
                    "Decoding error while processing file %s: %s" %
                    (filename1, str(err)))

            lineno += delta_lineno
            try:
                line2 = read_and_filter_line(INPUT2, ignore, filter)[0]
            except UnicodeDecodeError:
                err = sys.exc_info()[1]
                raise RuntimeError(
                    "Decoding error while processing file %s: %s" %
                    (filename2, str(err)))

            #print "line1 '%s'" % line1
            #print "line2 '%s'" % line2

            if line1 is None and line2 is None:
                return [False, None, ""]

            if line1 is None or line2 is None:
                return [True, lineno, file_diff(
                    filename1, filename2, lineno=lineno)]

            try:
                floats1, line1 = _extract_floats(line1, float_p)
                floats2, line2 = _extract_floats(line2, float_p)
            except:
                return [True, lineno, file_diff(
                    filename1, filename2, lineno=lineno)]

            #print "floats1 '%s'" % floats1
            #print "floats2 '%s'" % floats2

            if len(floats1) != len(floats2):
                return [True, lineno, file_diff(
                    filename1, filename2, lineno=lineno)]

            for v1, v2 in zip(floats1, floats2):
                vDiff = math.fabs(v1 - v2)
                vMax = max(math.fabs(v1), math.fabs(v2))
                if vDiff > absolute_tolerance and \
                   vDiff / vMax > relative_tolerance:
                    return [True, lineno, file_diff(
                        filename1, filename2, lineno=lineno)]

            line1 = whitespace_p.sub(' ', line1.strip())
            line2 = whitespace_p.sub(' ', line2.strip())

            #print "Line1 '%s'" % line1
            #print "Line2 '%s'" % line2

            if line1 != line2:
                return [True, lineno, file_diff(
                    filename1, filename2, lineno=lineno)]
    finally:
        if INPUT1 is not None:
            INPUT1.close()
        if INPUT2 is not None:
            INPUT2.close()

    return [False, None, ""]


def compare_file(filename1,
                 filename2,
                 ignore=["\t", " ", "\n", "\r"],
                 filter=None,
                 tolerance=None):
    """
    Do a simple comparison of two files that ignores differences
    in newline types.  If filename1 or filename2 is a zipfile, then it is
    assumed to contain a single file.

    The return value is the tuple: (status,lineno).  If status is True,
    then a difference has occured on the specified line number.  If
    the status is False, then lineno is None.

    The goal of this utility is to simply indicate whether there are
    differences in files.  The Python 'difflib' is much more comprehensive
    and consequently more costly to apply.  The shutil.filecmp utility is
    similar, but it does not ignore differences in file newlines.  Also,
    this utility can ignore an arbitrary set of characters.

    The 'filter' function evaluates each line separately.  If it returns True,
    then that line should be ignored.  If it returns a string, then that string replaces
    the line.
    """
    if tolerance is not None:
        tmp = copy.copy(ignore)
        tmp.remove(' ')
        tmp.remove('\t')
        try:
            tol, strict = tolerance
        except:
            tol = tolerance
            strict = True
        return compare_file_with_numeric_values(
            filename1,
            filename2,
            ignore=tmp,
            filter=filter,
            tolerance=tol,
            strict_numbers=strict)

    if not os.path.exists(filename1):
        raise IOError("compare_file: cannot find file `" + filename1 + "' (in "
                      + os.getcwd() + ")")
    if not os.path.exists(filename2):
        raise IOError("compare_file: cannot find file `" + filename2 + "' (in "
                      + os.getcwd() + ")")

    INPUT1 = INPUT2 = None
    try:
        INPUT1 = open_possibly_compressed_file(filename1)
        INPUT2 = open_possibly_compressed_file(filename2)
        #
        # This is check is deferred until the zipfiles are setup to ensure a
        # consistent logic for zipfile analysis.  If the files are the same,
        # but they are zipfiles with > 1 files, then we raise an exception.
        #
        if not sys.platform.startswith('win') and os.stat(filename1) == os.stat(
                filename2):
            return [False, None, ""]
        #
        lineno = 0
        while True:
            # If either line is composed entirely of characters to
            # ignore, then get another one.  In this way we can
            # skip blank lines that are in one file but not the other

            line1, delta_lineno = read_and_filter_line(INPUT1, ignore, filter)
            lineno += delta_lineno
            line2 = read_and_filter_line(INPUT2, ignore, filter)[0]

            if line1 is None and line2 is None:
                return [False, None, ""]

            if line1 != line2:
                return [True, lineno, file_diff(
                    filename1, filename2, lineno=lineno)]
    finally:
        if INPUT1 is not None:
            INPUT1.close()
        if INPUT2 is not None:
            INPUT2.close()


def compare_large_file(filename1,
                       filename2,
                       ignore=["\t", " ", "\n", "\r"],
                       bufSize=1 * 1024 * 1024):
    """
    Do a simple comparison of two files that ignores white space, or
    characters specified in "ignore" list.

    The return value is True if a difference is found, False otherwise.

    For very long text files, this function will be faster than
    compare_file() because it reads the files in by large chunks
    instead of by line.  The cost is that you don't get the lineno
    at which the difference occurs.
    """

    INPUT1 = open_possibly_compressed_file(filename1)
    try:
        INPUT2 = open_possibly_compressed_file(filename2)
    except IOError:
        INPUT1.close()
        raise
    #
    # This is check is deferred until the zipfiles are setup to ensure a consistent logic for
    # zipfile analysis.  If the files are the same, but they are zipfiles with > 1 files, then we
    # raise an exception.
    #
    if not sys.platform.startswith('win') and os.stat(filename1) == os.stat(
            filename2):
        INPUT1.close()
        INPUT2.close()
        return False

    f1Size = os.stat(filename1).st_size
    f2Size = os.stat(filename2).st_size

    result = False

    while True:
        buf1 = get_desired_chars_from_file(INPUT1, bufSize, ignore)
        buf2 = get_desired_chars_from_file(INPUT2, bufSize, ignore)

        if len(buf1) == 0 and len(buf2) == 0:
            break
        elif len(buf1) == 0 or len(buf2) == 0:
            result = True
            break

        if len(buf1) != len(buf2) or buf1 != buf2:
            result = True
            break

    INPUT1.close()
    INPUT2.close()
    return result



import os
from os.path import abspath, dirname
currdir = dirname(abspath(__file__)) + os.sep

import pyutilib.th as unittest


class TestNoCategory(unittest.TestCase):
    def test_noCategory(self):
        self.assertRaises(
            AttributeError, getattr, self.test_noCategory, 'smoke')
        self.assertRaises(
            AttributeError, getattr, self.test_noCategory, 'nightly')
        self.assertRaises(
            AttributeError, getattr, self.test_noCategory, 'expensive')
        self.assertRaises(
            AttributeError, getattr, self.test_noCategory, 'fragile')
        self.assertRaises(
            AttributeError, getattr, self.test_noCategory, 'custom')

    @unittest.category('smoke')
    def test_smoke(self):
        self.assertEqual(self.test_smoke.smoke, 1)
        self.assertEqual(self.test_smoke.nightly, 0)
        self.assertEqual(self.test_smoke.expensive, 0)
        self.assertEqual(self.test_smoke.fragile, 0)
        self.assertRaises(
            AttributeError, getattr, self.test_smoke, 'custom')

    @unittest.category('expensive')
    def test_expensive(self):
        self.assertEqual(self.test_expensive.smoke, 0)
        self.assertEqual(self.test_expensive.nightly, 0)
        self.assertEqual(self.test_expensive.expensive, 1)
        self.assertEqual(self.test_expensive.fragile, 0)
        self.assertRaises(
            AttributeError, getattr, self.test_expensive, 'custom')

    @unittest.category('!expensive')
    def test_notExpensive(self):
        self.assertEqual(self.test_notExpensive.smoke, 0)
        self.assertEqual(self.test_notExpensive.nightly, 0)
        self.assertEqual(self.test_notExpensive.expensive, 0)
        self.assertEqual(self.test_notExpensive.fragile, 0)
        self.assertRaises(
            AttributeError, getattr, self.test_notExpensive, 'custom')

    @unittest.category('fragile')
    def test_fragile(self):
        self.assertRaises(
            AttributeError, getattr, self.test_noCategory, 'smoke')
        self.assertRaises(
            AttributeError, getattr, self.test_noCategory, 'nightly')
        self.assertRaises(
            AttributeError, getattr, self.test_noCategory, 'expensive')
        self.assertEqual(self.test_fragile.fragile, 1)
        self.assertRaises(
            AttributeError, getattr, self.test_notExpensive, 'custom')

    @unittest.category('fragile','smoke')
    def test_fragile_smoke(self):
        self.assertEqual(self.test_fragile_smoke.smoke, 1)
        self.assertEqual(self.test_fragile_smoke.nightly, 0)
        self.assertEqual(self.test_fragile_smoke.expensive, 0)
        self.assertEqual(self.test_fragile_smoke.fragile, 1)
        self.assertRaises(
            AttributeError, getattr, self.test_notExpensive, 'custom')

    @unittest.category('custom')
    def test_custom(self):
        self.assertEqual(self.test_custom.smoke, 0)
        self.assertEqual(self.test_custom.nightly, 0)
        self.assertEqual(self.test_custom.expensive, 0)
        self.assertEqual(self.test_custom.fragile, 0)
        self.assertEqual(self.test_custom.custom, 1)

    @unittest.category('custom','smoke')
    def test_multi(self):
        self.assertEqual(self.test_multi.smoke, 1)
        self.assertEqual(self.test_multi.nightly, 0)
        self.assertEqual(self.test_multi.expensive, 0)
        self.assertEqual(self.test_multi.fragile, 0)
        self.assertEqual(self.test_multi.custom, 1)

@unittest.category('expensive')
class TestExpensive(unittest.TestCase):
    def test_noCategory(self):
        self.assertRaises(
            AttributeError, getattr, self.test_noCategory, 'smoke')
        self.assertRaises(
            AttributeError, getattr, self.test_noCategory, 'nightly')
        self.assertRaises(
            AttributeError, getattr, self.test_noCategory, 'expensive')
        self.assertRaises(
            AttributeError, getattr, self.test_noCategory, 'fragile')
        self.assertRaises(
            AttributeError, getattr, self.test_noCategory, 'custom')

    @unittest.category('smoke')
    def test_smoke(self):
        self.assertEqual(self.test_smoke.smoke, 1)
        self.assertEqual(self.test_smoke.nightly, 0)
        self.assertEqual(self.test_smoke.expensive, 1)
        self.assertEqual(self.test_smoke.fragile, 0)
        self.assertRaises(
            AttributeError, getattr, self.test_smoke, 'custom')

    @unittest.category('expensive')
    def test_expensive(self):
        self.assertEqual(self.test_expensive.smoke, 0)
        self.assertEqual(self.test_expensive.nightly, 0)
        self.assertEqual(self.test_expensive.expensive, 1)
        self.assertEqual(self.test_expensive.fragile, 0)
        self.assertRaises(
            AttributeError, getattr, self.test_expensive, 'custom')

    @unittest.category('!expensive')
    def test_notExpensive(self):
        self.assertEqual(self.test_notExpensive.smoke, 0)
        self.assertEqual(self.test_notExpensive.nightly, 0)
        self.assertEqual(self.test_notExpensive.expensive, 0)
        self.assertEqual(self.test_notExpensive.fragile, 0)
        self.assertRaises(
            AttributeError, getattr, self.test_notExpensive, 'custom')

    @unittest.category('fragile')
    def test_fragile(self):
        self.assertEqual(self.test_fragile.smoke, 0)
        self.assertEqual(self.test_fragile.nightly, 0)
        self.assertEqual(self.test_fragile.expensive, 1)
        self.assertEqual(self.test_fragile.fragile, 1)
        self.assertRaises(
            AttributeError, getattr, self.test_notExpensive, 'custom')

    @unittest.category('fragile','smoke')
    def test_fragile_smoke(self):
        self.assertEqual(self.test_fragile_smoke.smoke, 1)
        self.assertEqual(self.test_fragile_smoke.nightly, 0)
        self.assertEqual(self.test_fragile_smoke.expensive, 1)
        self.assertEqual(self.test_fragile_smoke.fragile, 1)
        self.assertRaises(
            AttributeError, getattr, self.test_notExpensive, 'custom')

    @unittest.category('custom')
    def test_custom(self):
        self.assertEqual(self.test_custom.smoke, 0)
        self.assertEqual(self.test_custom.nightly, 0)
        self.assertEqual(self.test_custom.expensive, 1)
        self.assertEqual(self.test_custom.fragile, 0)
        self.assertEqual(self.test_custom.custom, 1)

    @unittest.category('custom','smoke')
    def test_multi(self):
        self.assertEqual(self.test_multi.smoke, 1)
        self.assertEqual(self.test_multi.nightly, 0)
        self.assertEqual(self.test_multi.expensive, 1)
        self.assertEqual(self.test_multi.fragile, 0)
        self.assertEqual(self.test_multi.custom, 1)


if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
#

try:
    import yaml
    using_yaml = True
except ImportError:  #pragma:nocover
    using_yaml = False
from pyutilib.component.core import SingletonPlugin, implements
from pyutilib.autotest import plugins
import pyutilib.misc


class YamlTestParser(SingletonPlugin):

    implements(plugins.ITestParser)

    def __init__(self, **kwds):
        SingletonPlugin.__init__(self, **kwds)
        self.name = 'yml'

    def load_test_config(self, filename):
        if using_yaml:
            INPUT = open(filename, 'r')
            repn = yaml.load(INPUT, yaml.SafeLoader)
            INPUT.close()
            return repn
        #
        # If PyYaml is not available, then we use a simple yaml parser
        #
        INPUT = open(filename, 'r')
        repn = pyutilib.misc.simple_yaml_parser(INPUT)
        INPUT.close()
        return repn

    def print_test_config(self, repn):
        print(repn)  #pragma:nocover

    def enabled(self):
        return True



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ['Resource']

from pyutilib.workflow import globals


class Resource(object):

    def __init__(self, id=None, name=None):
        if not id is None:
            # TBD:  does this ever happen?
            self.id = id
        else:
            self.id = globals.unique_id()
        if name is None:
            self.description = "Resource" + str(self.id)
            self.name = self.description
        else:
            # TBD:  does this ever happen?
            self.description = name
            self.name = name
        self._busy = False

    def available(self):
        return not self._busy  #pragma:nocover

    def lock(self):
        self._busy = True  #pragma:nocover

    def unlock(self):
        self._busy = False  #pragma:nocover

    def __repr__(self):
        return str(self)  #pragma:nocover

    def __str__(self):
        return "%s" % str(self.name)  #pragma:nocover



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

from pyutilib.th.pyunit import TestCase, TestResult, TestSuite, TextTestRunner, main, nottest, category
try:
    from pyutilib.th.pyunit import skip, skipIf, skipUnless, expectedFailure, SkipTest
except:
    pass

try:
    import pyutilib.th.nose_testdata
except ImportError:
    pass
try:
    import pyutilib.th.nose_gc
except ImportError:
    pass
try:
    import pyutilib.th.nose_timeout
except ImportError:
    pass



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ['registered_executable', 'register_executable', 'TempfileManager']

from pyutilib.component.core import ExtensionPoint, PluginGlobals
from pyutilib.component.config import TempfileManager
from pyutilib.component.executables import IExternalExecutable, ExternalExecutable
"""
Test if an exectuable is registered, using the IExternalExecutable extension
point.

If 'name' is None, then return a list of the names of all registered
executables that are enabled.

If either this executable is not registered or it is disabled, then
None is returned.
"""


def registered_executable(name=None):
    ep = ExtensionPoint(IExternalExecutable)
    if name is None:
        return filter(lambda x: x.name, ep.extensions())
    return ep.service(name)


"""
Register an executable, using the IExternalExecutable extension
point.

If this executable has been registered, then do not reregister it
(even if it is disabled).
"""


def register_executable(name, validate=None):
    ep = ExtensionPoint(IExternalExecutable)
    if len(ep(name, all=True)) == 0:
        PluginGlobals.add_env("pca")
        PluginGlobals._executables.append(
            ExternalExecutable(
                name=name, validate=validate))
        PluginGlobals.pop_env()
    else:
        #
        # If the executable is being 'registered', then we search for it
        # again, since the user environment may have changed.
        #
        list(ep(name, all=True))[0].find_executable()



import logging
import os
from os.path import abspath, dirname
currdir = dirname(abspath(__file__)) + os.sep

import pyutilib.th as unittest
import pyutilib.misc
from pyutilib.workflow import *

try:
    import json
    json_available = True
except:
    json_available = False


class Handler(logging.StreamHandler):

    def emit(self, record):
        raise RuntimeError(str(record))


handler = Handler()
logger = logging.getLogger('pyutilib.workflow')


class TestData(unittest.TestCase):

    def test1(self):
        # Print FunctorAPIData string
        data = FunctorAPIData()
        data.a = 1
        data.b = [1, 2]
        data.c = FunctorAPIData()
        data.c.x = 1
        data.c.y = 2
        data.aa = 'here is more'
        data.clean()
        pyutilib.misc.setup_redirect(currdir + 'test1.out')
        print(data)
        pyutilib.misc.reset_redirect()
        self.assertFileEqualsBaseline(currdir + 'test1.out',
                                      currdir + 'test1.txt')
        self.assertEquals(len(data._dirty_), 0)

    @unittest.skipIf(not json_available, "JSON not available")
    def test2(self):
        # Print FunctorAPIData representation
        data = FunctorAPIData()
        data.a = 1
        data.b = [1, 2]
        data.c = FunctorAPIData()
        data.c.x = 1
        data.c.y = 2
        data['aa'] = 'here is more'
        data.clean()
        FILE = open(currdir + 'test2.out', 'w')
        json.dump(data, FILE)
        FILE.close()
        self.assertMatchesJsonBaseline(currdir + 'test2.out',
                                       currdir + 'test2.txt')
        self.assertEquals(len(data._dirty_), 0)

    @unittest.expectedFailure
    def test_err1(self):
        # Unknown attribute
        data = FunctorAPIData()
        data._x

    @unittest.expectedFailure
    def test_err2(self):
        # Undeclared attribute
        data = FunctorAPIData()
        data.declare('a')
        data.x

    @unittest.expectedFailure
    def test_err3(self):
        # Undeclared attribute
        data = FunctorAPIData()
        data.declare(['a'])
        data.x


class TestAPI(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Disable the pyutilib.workflow logging handler
        if len(logger.handlers) > 0:
            cls._handler = logger.handlers[0]
            logger.removeHandler(cls._handler)
        else:
            cls._handler = None
        logger.addHandler(handler)

    @classmethod
    def tearDownClass(cls):
        # Re-enable the pyutilib.workflow logging handler
        logger.removeHandler(handler)
        if not cls._handler is None:
            logger.addHandler(cls._handler)

    def test1(self):
        # Simple test: no keyword arguments or return values
        @functor_api
        def test1(data):
            # Required:
            #     data: input data
            data.a = 2
            data.b[0] = 2
        #
        options = FunctorAPIData()
        options.a = 1
        options.b = [1, 2]
        retval = test1(options)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test1(data=options)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        #
        self.assertTrue('test1' in FunctorAPIFactory.services())

    def test1a(self):
        """Simple test: no keyword arguments or return values"""

        @functor_api
        def test1a(data):
            """
            Required:
                data: input data
            Return:
                data: output data
            """
            data.a = 2
            data.b[0] = 2
        #
        options = FunctorAPIData()
        options.a = 1
        options.b = [1, 2]
        retval = test1a(options)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test1a(data=options)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])

    def test1b(self):
        """Simple test: data keyword argument, no return values"""

        @functor_api
        def test1b(data=None):
            data.a = 2
            data.b[0] = 2
        #
        options = FunctorAPIData()
        options.a = 1
        options.b = [1, 2]
        retval = test1b(options)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test1b(data=options)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])

    def test2(self):
        """Simple test: no keyword arguments, returning data"""

        @functor_api
        def test2(data):
            data.a = 2
            data.b[0] = 2
            return data
        #
        options = FunctorAPIData()
        options.a = 1
        options.b = [1, 2]
        retval = test2(options)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test2(data=options)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])

    def test2a(self):
        """Simple test: no keyword arguments, returning data"""

        @functor_api
        def test2a(data):
            """
            Required:
                data: input data
            Return:
                data: output data
            """
            data.a = 2
            data.b[0] = 2
            return data
        #
        options = FunctorAPIData()
        options.a = 1
        options.b = [1, 2]
        retval = test2a(options)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test2a(data=options)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])

    def test3(self):
        """Simple test: keyword arguments, no return values"""

        @functor_api
        def test3(data, x=1, y=2):
            """
            Required:
                data: input data
                x: integer
            Optional:
                y: integer
            """
            data.a = y
            data.b[0] = x
        #
        options = FunctorAPIData()
        options.a = 1
        options.b = [1, 2]
        retval = test3(options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test3(data=options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])

    def test3a(self):
        """Simple test: keyword arguments, no return values"""

        @functor_api
        def test3a(x=1, y=2, data=None):
            """
            Required:
                data: input data
                x: integer
            Optional:
                y: integer
            """
            data.a = y
            data.b[0] = x
        #
        options = FunctorAPIData()
        options.a = 1
        options.b = [1, 2]
        retval = test3a(options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test3a(data=options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])

    def test4(self):
        """Simple test: keyword arguments, simple return value"""

        @functor_api
        def test4(data, x=1, y=2):
            """
            Required:
                data: input data
                x: integer
            Optional:
                y: integer
            Return:
                data: output data
            """
            data.a = y
            data.b[0] = x
            return data
        #
        options = FunctorAPIData()
        options.a = 1
        options.b = [1, 2]
        retval = test4(options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test4(data=options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])

    def test5(self):
        """Simple test: keyword arguments, non-data return values"""

        @functor_api(outputs=('z'))
        def test5(data, x=1, y=2):
            """
            Required:
                data: input data
                x: integer
            Optional:
                y: integer
            Return:
                data: output data
                z: integer
            """
            data.a = y
            data.b[0] = x
            return FunctorAPIData(z=x)
        #
        options = FunctorAPIData()
        options.a = 1
        options.b = [1, 2]
        retval = test5(options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test5(data=options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        self.assertEquals(retval.z, 2)

    def test6(self):
        """Simple test: keyword arguments, non-data return values with data"""

        @functor_api(outputs=('z'))
        def test6(data, x=1, y=2):
            """
            Required:
                data: input data
                x: integer
            Optional:
                y: integer
            Return:
                data: output data
                z: integer
            """
            data.a = y
            data.b[0] = x
            return FunctorAPIData(data=data, z=x)
        #
        options = FunctorAPIData()
        options.a = 1
        options.b = [1, 2]
        retval = test6(options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test6(data=options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        self.assertEquals(retval.z, 2)

    def test5a(self):
        """Outputs specified in docstring: keyword arguments, non-data return values"""

        @functor_api
        def test5a(data, x=1, y=2):
            """
            Required:
                data: input data
                x: integer
            Optional:
                y: integer
            Return:
                data: output data
                z: integer
            """
            data.a = y
            data.b[0] = x
            return FunctorAPIData(z=x)
        #
        options = FunctorAPIData()
        options.a = 1
        options.b = [1, 2]
        retval = test5a(options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test5a(data=options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        self.assertEquals(retval.z, 2)

    def test6a(self):
        """Outputs specified in docstring: keyword arguments, non-data return values with data"""

        @functor_api
        def test6a(data, x=1, y=2):
            """
            Required:
                data: input data
                x: integer
            Optional:
                y: integer
            Return:
                data: output data
                z: integer
            """
            data.a = y
            data.b[0] = x
            return FunctorAPIData(data=data, z=x)
        #
        options = FunctorAPIData()
        options.a = 1
        options.b = [1, 2]
        retval = test6a(options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test6a(data=options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        self.assertEquals(retval.z, 2)

    def test7a(self):
        """Test with dict data"""

        @functor_api
        def test7a(data, x=1, y=2):
            """
            Required:
                data: input data
                x: integer
            Optional:
                y: integer
            """
            data.a = y
            data.b[0] = x
        #
        options = {}
        options['a'] = 1
        options['b'] = [1, 2]
        retval = test7a(options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test7a(data=options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])

    def test7b(self):
        """Test with dict data and return a dictionary"""

        @functor_api
        def test7b(data, x=1, y=2):
            """
            Required:
                data: input data
                x: integer
            Optional:
                y: integer
            """
            data.a = y
            data.b[0] = x
            return {'data': data}
        #
        options = {}
        options['a'] = 1
        options['b'] = [1, 2]
        retval = test7b(options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test7b(data=options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])

    def test7c(self):
        """Test with dict data and return a dictionary"""

        @functor_api
        def test7c(data, x=1, y=2):
            """
            Required:
                data: input data
                x: integer
            Optional:
                y: integer
            Return:
                z: integer
            """
            data.a = y
            data.b[0] = x
            return {'data': data, 'z': x}
        #
        options = {}
        options['a'] = 1
        options['b'] = [1, 2]
        retval = test7c(options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test7c(data=options, x=2)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        self.assertEquals(retval.z, 2)

    def test8(self):
        """Simple test with required nested data"""

        @functor_api
        def test8(data):
            """
            Required:
                data: input data
                data.foo.bar:
            """
            data.foo.foo = 3
            #data.a = 2
            #data.b[0] = 2
        #

        options = FunctorAPIData()
        options.foo = FunctorAPIData()
        options.foo.bar = 1
        options.a = 1
        options.b = [1, 2]
        retval = test8(options)
        self.assertEquals(retval.data.a, 1)
        self.assertEquals(retval.data.b, [1, 2])
        self.assertEquals(retval.data.foo.foo, 3)
        #retval = test8(data=options)
        #self.assertEquals(retval.data.a, 2)
        #self.assertEquals(retval.data.b, [2,2])

    def test9(self):
        """Simple test: no keyword arguments or return values"""

        @functor_api
        def test9(data):
            """
            This is the
            short
            documentation.

            This

            is
  
            the
  
            long documentation.
            Required:
                data: multiline
                      description of 
                      input data object
                data.a: More data
            """
            data.a = 2
            data.b[0] = 2
        #
        options = FunctorAPIData()
        options.a = 1
        options.b = [1, 2]
        retval = test9(options)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        retval = test9(data=options)
        self.assertEquals(retval.data.a, 2)
        self.assertEquals(retval.data.b, [2, 2])
        #
        self.assertTrue('test1' in FunctorAPIFactory.services())
        self.assertEquals(test9.__short_doc__,
                          'This is the\nshort\ndocumentation.')
        self.assertEquals(test9.__long_doc__,
                          'This\n\nis\n\nthe\n\nlong documentation.')

    @unittest.expectedFailure
    def test10(self):
        """Simple test: no keyword arguments or return values"""

        @functor_api
        def test10(data=None, x=1):
            """
            Required:
                x: 
            Optional:
                data:
            """
            return FunctorAPIData(z=2 * x)
        #
        retval = test10(x=3)
        self.assertEquals(retval.z, 6)

    @unittest.expectedFailure
    def test_err1(self):
        """Expect an error when variable length arguments are supported"""

        @functor_api
        def err1(*args):
            pass

    @unittest.expectedFailure
    def test_err2(self):
        """Expect an error when variable length keyword arguments are supported"""

        @functor_api
        def err2(**kwargs):
            pass

    @unittest.expectedFailure
    def test_err3(self):
        """Expect an error when return value is not None or Options()"""

        @functor_api
        def err3(data):
            return 1

        f(FunctorAPIData())

    @unittest.expectedFailure
    def test_err4(self):
        """Expect an error when no data argument is specified"""

        @functor_api
        def err4(data):
            data.a = 2
            data.b[0] = 2

        test1()

    @unittest.expectedFailure
    def test_err5(self):
        """Expect an error when an unspecified return value is given"""

        @functor_api
        def err5(data):
            return FunctorAPIData(z=None)

        f(FunctorAPIData())

    @unittest.expectedFailure
    def test_err6(self):
        """Expect an error when no data argument is specified"""

        @functor_api
        def err6():
            pass

        f(FunctorAPIData())

    @unittest.expectedFailure
    def test_err7a(self):
        """Argument missing from docstring"""

        @functor_api
        def err7a(data, x=1, y=2):
            """
            Optional:
                y: integer
            """
            pass

    @unittest.expectedFailure
    def test_err7b(self):
        """Argument missing from docstring"""

        @functor_api
        def err7b(data, x=1, y=2):
            """
            Required:
                x: integer
            """
            pass

    @unittest.expectedFailure
    def test_err7c(self):
        """Argument missing from docstring"""

        @functor_api(outputs=('z'))
        def err7c(data, x=1, y=2):
            """
            Required:
                x: integer
            Optional:
                y: integer
            """
            return FunctorAPIData(z=1)

    @unittest.expectedFailure
    def test_err7A(self):
        """Unexpected required argument"""

        @functor_api
        def err7A(data, x=1, y=2):
            """
            Required:
                x: integer
                bad: integer
            Optional:
                y: integer
            """
            pass

    @unittest.expectedFailure
    def test_err7B(self):
        """Argument missing from docstring"""

        @functor_api
        def err7B(data, x=1, y=2):
            """
            Required:
                x: integer
            Optional:
                y: integer
                bad: integer
            """
            pass

    @unittest.expectedFailure
    def test_err7C(self):
        """Argument missing from docstring"""

        @functor_api(outputs=('z'))
        def err7C(data, x=1, y=2):
            """
            Required:
                x: integer
            Optional:
                y: integer
            Return:
                z: integer
                bad: integer
            """
            return FunctorAPIData(z=1)

    @unittest.expectedFailure
    def test_err8a(self):
        """Missing nested value"""

        @functor_api
        def err8a(data):
            """
            Required:
                data.x: integer
            """
            pass

        err8a(FunctorAPIData())

    @unittest.expectedFailure
    def test_err8b(self):
        """Nested value with None value"""

        @functor_api
        def err8b(data):
            """
            Required:
                data.x: integer
            """
            pass

        err8b(FunctorAPIData(data=FunctorAPIData()))

    @unittest.expectedFailure
    def test_err8c(self):
        """Nested value with None value"""

        @functor_api
        def err8c(data):
            """
            Required:
                data.x.y: integer
            """
            pass

        err8c(FunctorAPIData(x={}))

    @unittest.expectedFailure
    def test_err9(self):
        """Redefinied test functions"""

        @functor_api
        def err9(data):
            pass

        @functor_api
        def err9(data):
            pass

    @unittest.expectedFailure
    def test_err10(self):
        """Simple test with required nested data"""

        @functor_api
        def err10(data):
            """
            Required:
                data: input data
                data.foo.bar:
            """
            data.foo.foo = 3
            data.a = 2
        #
        options = FunctorAPIData()
        options.foo = FunctorAPIData()
        options.foo.bar = 1
        options.a = 1
        options.b = [1, 2]
        retval = err10(options)
        self.assertEquals(retval.data.a, 1)
        self.assertEquals(retval.data.b, [1, 2])
        self.assertEquals(retval.data.foo.foo, 3)

    @unittest.expectedFailure
    def test_err10a(self):
        """Expect an error when the same functor is defined twice"""

        @functor_api
        def err10a(data):
            pass

        @functor_api
        def err10a(data):
            pass

    @unittest.expectedFailure
    def test_err10b(self):
        """Expect an error when the same functor is defined twice"""

        @functor_api(namespace='foo')
        def err10b(data):
            pass

        @functor_api(namespace='foo')
        def err10b(data):
            pass

    @unittest.expectedFailure
    def test_err11(self):
        """Expect an error when 'data' is not defined when it is required"""

        @functor_api
        def err11(data=None, x=1):
            """
            Required:
                data: 
                x: 
            """

        err11(x=2)

    @unittest.expectedFailure
    def test_err12(self):
        """Expect an error when multiple data options are provided"""

        @functor_api
        def err12(data):
            pass

        err12({}, {})

    @unittest.expectedFailure
    def test_err13(self):
        """Expect an error when returning something other than None, FunctorAPIData or a dict object"""

        @functor_api
        def err13(data):
            return set()

        err13({})


if __name__ == "__main__":
    unittest.main()



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
"""
This package defines the Configuration class, which provides a generic interface for reading/writing configuration files.  This class uses a simple
model for configuration data:

  section -> option -> value

A key role of this class is to support initialization of Option objects.
"""

import os.path
from pyutilib.component.core import PluginError, Interface, Plugin, ExtensionPoint, implements, IOptionDataProvider
from pyutilib.component.config.options import IOption, IFileOption, IUpdatedOptionsAction


class ConfigurationError(PluginError):
    """Exception raised when a value in the configuration file is not valid."""


class IConfiguration(Interface):
    """Define an interface for classes that are used by Configuration to
    read/write configuration data."""

    def load(self, filename):
        """Returns a list of tuples: [ (section,option,value) ]"""

    def save(self, filename, config):
        """Save configuration information to the specified file."""


class Configuration(Plugin):
    """This class manages configuration data.  Further, this configuration
    I/O is coordinated with Option objects.  When configuration data is read
    in, associated Option plugins are populated.  Similarly, when
    configuration data is writen, the configuration data is taken from
    Option data."""

    def __init__(self, filename=None, parser="ConfigParser"):
        """Constructor.
            @param filename - The associated configuration file.
            @param parser   - Specify the name of the parser used to
                read/write configuration files.
        """
        self.parser_type = "Configuration_ConfigParser"
        self.filename = filename
        #
        # Define extension points
        #
        self.parsers = ExtensionPoint(IConfiguration)
        self.option_plugins = ExtensionPoint(IOption)
        self.option_data_plugin = ExtensionPoint(IOptionDataProvider)
        self.pathoption_plugins = ExtensionPoint(IFileOption)
        self.postconfig_actions = ExtensionPoint(IUpdatedOptionsAction)
        self.clear()

    def clear(self):
        """Clear local data."""
        self.config = []
        self.data = {}
        self.section = []

    def __contains__(self, name):
        """Return whether the configuration contains a section of the given
        name.
        """
        return name in self.data

    def __getitem__(self, name):
        """Return the configuration section with the specified name."""
        if name not in self.data:
            raise ConfigurationError("No section " + name + " in data")
        return self.data[name]

    def sections(self):
        """Returns the names of all sections in the configuration data."""
        return list(self.data.keys())

    def load(self, filename=None):
        """Load configuration from a file."""
        if len(self.parsers) == 0:  #pragma:nocover
            raise ConfigurationError("No IConfiguration parsers are registered")
        if not filename is None:
            self.filename = filename
        if self.filename is None:
            raise ConfigurationError("Cannot load without a filename")
        for option in self.pathoption_plugins:
            option.set_dir(os.path.dirname(self.filename))
        #
        # By default, we simply take the first parser
        #
        self.config = self.parsers.service(self.parser_type).load(self.filename)
        self.data = {}
        self.section = []
        for (s, o, v) in self.config:
            if not s in self.data:
                self.section.append(s)
                self.data[s] = {}
            if not o in self.data[s]:
                self.data[s][o] = []
            self.data[s][o].append(v)
        #
        # Iterate through all sections, in the order they were
        # loaded.  Load data for extensions that match each section name.
        #
        for sec in self.section:
            #
            # Find the option_plugins that match this section
            #
            plugins = []
            for plugin in self.option_plugins:
                if plugin.matches_section(sec):
                    plugins.append(plugin)
            for option in self.data[sec]:
                flag = False
                for plugin in plugins:
                    if plugin.matches_name(option):
                        flag = plugin.load(option,
                                           self.data[sec][option]) or flag
                if not flag:
                    raise ConfigurationError(
                        "Problem loading file %r. Option %r in section %r is not recognized!"
                        % (self.filename, option, sec))
        #
        # Finalize the configuration process
        #
        for plugin in self.postconfig_actions:
            plugin.reset_after_updates()

    def save(self, filename=None):
        """Save configuration to a file."""
        if not filename is None:
            self.filename = filename
        if self.filename is None:
            raise ConfigurationError("Cannot save without a filename")
        #
        # Setup the list of tuples
        #
        self.clear()
        self.data = self.option_data_plugin.service().get_data()
        self.section = list(self.data.keys())
        self.section.sort()
        flag = False
        header = "\nNote: the following configuration options have been omitted because their\nvalue is 'None':\n"
        for sec in self.section:
            plugins = []
            for plugin in self.option_plugins:
                if plugin.matches_section(sec):
                    plugins.append(plugin)
            #
            options = list(self.data[sec].keys())
            options.sort()
            for option in options:
                for plugin in plugins:
                    if plugin.matches_name(option):
                        if not self.data[sec][option] is None:
                            val = self.data[sec][option]
                            self.config.append((sec, option, val))
                        else:
                            flag = True
                            header = header + "  section=%r option=%r\n" % (
                                sec, option)
                        break
        if flag:
            header = header + "\n"
        else:
            header = None
        #
        # Write config file
        #
        self.parsers.service(self.parser_type).save(self.filename, self.config,
                                                    header)

    def pprint(self):
        """Print a simple summary of the configuration data."""
        text = ""
        for (s, o, v) in self.config:
            text += "[%s] %s = %s\n" % (s, o, v)
        print(text)

    def summarize(self):
        """Summarize options"""
        tmp = {}
        for option in self.option_plugins:
            tmp.setdefault(option.section, {})[option.name] = option
        keys = list(tmp.keys())
        keys.sort()
        for key in keys:
            print("[" + key + "]")
            print("")
            okeys = list(tmp[key].keys())
            okeys.sort()
            for okey in okeys:
                print("  Option:    " + tmp[key][okey].name)
                print("  Type:      " + tmp[key][okey].__class__.__name__)
                print("  Default:   " + tmp[key][okey].default_str())
                print("  Doc:       " + tmp[key][okey].__doc__)
                print("")
            print("")



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________

__all__ = ['infinity', 'nan', 'is_nan', 'is_finite']
"""
Definitions of mathematical constants
"""

import sys
import math

if sys.version_info < (2, 6):
    """ Definition of infinity """
    infinity = float(1e3000)
    """ Definition of NaN """
    nan = infinity / infinity

    def is_nan(x):
        """
        Returns true if the argument is a float and it does not equal itself
        """
        return type(x) is float and x != x

    def is_finite(val):
        """
        Returns true if the argument is a float or int and it is not infinite or NaN
        """
        return type(val) in (float, int) and val not in (infinity, -infinity,
                                                         nan)

else:
    """ Definition of infinity """
    infinity = float('inf')
    """ Definition of NaN """
    nan = infinity / infinity

    def is_nan(x):
        """
        Returns true if the argument is a float and it does not equal itself
        """
        try:
            return math.isnan(x)
        except TypeError:
            return False

    def is_finite(x):
        """
        Returns true if the argument is a float or int and it is not infinite or NaN
        """
        try:
            return not math.isinf(x)
        except TypeError:
            return False



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
#

import sys
from pyutilib.component.core import Plugin, implements, alias
from pyutilib.autotest import plugins


class DefaultTestDriver(Plugin):
    """
    This is the 'default' test driver, which simply prints the
    arguments being passed into a test.
    """

    implements(plugins.ITestDriver)
    alias('default')

    def setUpClass(self, cls, options):
        """Set-up the class that defines the suite of tests"""

    def tearDownClass(self, cls, options):
        """Tear-down the class that defines the suite of tests"""

    def setUp(self, testcase, options):
        """Set-up a single test in the suite"""

    def tearDown(self, testcase, options):
        """Tear-down a single test in the suite"""
        sys.stdout.flush()

    def run_test(self, testcase, name, options):
        """Execute a single test in the suite"""
        print('run_test ' + name)
        print(options)



#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________


__all__ = ['register_executable', 'registered_executable', 'unregister_executable']

import pyutilib.misc
from pyutilib.factory.factory import CachedFactory

ExecutableFactory = CachedFactory('executables')


def unregister_executable(name):
    ExecutableFactory.unregister(name)


def register_executable(name, validate=None):
    #
    # Do not reregister an executable.
    #
    if name in ExecutableFactory:
        return

    @ExecutableFactory.register(name)
    class TMP(object):
        def __init__(self):
            self.executable = None
            self.exec_default = pyutilib.misc.search_file(
                name,
                implicitExt=pyutilib.misc.executable_extension,
                executable=True,
                validate=validate)

        def get_path(self):
            if self.executable:
                return self.executable
            return self.exec_default


def registered_executable(name=None):
    """
    Test if an exectuable is registered.

    If 'name' is None, then return a list of the names of all registered
    executables that are enabled.

    If this executable is not registered or it cannot be found, then
    None is returned.  Otherwise, return the path to the executable.

    NOTE: Since ExecutableFactory is cached, the search for the 
    executable only happens the first time the user calls registered_executable.
    """

    if name is None:
        return sorted(list(ExecutableFactory))
    if name in ExecutableFactory:
        tmp = ExecutableFactory(name)
        if tmp.get_path() is None:
            return None
        return tmp
    return None




#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
"""
Python modules that use Pyro to manage a generic task queue.

These modules were adapted from Pyro's distributed2 example.
"""

from pyutilib.pyro.util import get_nameserver, get_dispatchers, shutdown_pyro_components, using_pyro3, using_pyro4, Pyro
from pyutilib.pyro.task import Task, TaskProcessingError
from pyutilib.pyro.client import Client
from pyutilib.pyro.worker import TaskWorker, MultiTaskWorker, TaskWorkerServer
from pyutilib.pyro.dispatcher import Dispatcher, DispatcherServer
from pyutilib.pyro.nameserver import start_ns, start_nsc

#
# Pyro3 License
#

#
#Pyro Software License (MIT license):
#
#Pyro is Copyright (c) by Irmen de Jong.
#
#Permission is hereby granted, free of charge, to any person obtaining a
#copy of this software and associated documentation files (the "Software"),
#to deal in the Software without restriction, including without limitation
#the rights to use, copy, modify, merge, publish, distribute, sublicense,
#and/or sell copies of the Software, and to permit persons to whom the
#Software is furnished to do so, subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included
#in all copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
#OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
#ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
#OTHER DEALINGS IN THE SOFTWARE.
#
#This is the "MIT Software License" which is Open
#Source Initiative (OSI) certified, and GPL-compatible.  See
#http://www.opensource.org/licenses/mit-license.php (it strongly resembles
#the modified BSD license).
#

#
# Pyro4 License
#

#
#Pyro - Python Remote Objects
#Software License, copyright, and disclaimer
#
#  Pyro is Copyright (c) by Irmen de Jong (irmen@razorvine.net).
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#
#
#This is the "MIT Software License" which is OSI-certified, and GPL-compatible.
#See http://www.opensource.org/licenses/mit-license.php



#
# Unit Tests for util/*_io
#
#

import os
import sys
from os.path import abspath, dirname
pkgdir = dirname(abspath(__file__)) + os.sep + ".." + os.sep + ".."
currdir = dirname(abspath(__file__)) + os.sep
import six

import pyutilib.th as unittest
import pyutilib.misc


def filter1(str):
    return True


def filter2(str):
    return ""


def filter3(str):
    str = str.replace('THERE', 'HERE')
    str = str.replace('[4, 5, 6]', '[1,2,3]')
    return str


class IODebug(unittest.TestCase):

    def test_redirect1(self):
        # Verify that IO redirection works
        pyutilib.misc.setup_redirect(currdir + "redirect_io.out")
        print("HERE")
        print([1, 2, 3])
        #
        # Force a flush to ensure code coverage
        #
        sys.stdout.flush()
        pyutilib.misc.reset_redirect()
        self.assertFalse(
            pyutilib.misc.comparison.compare_file(
                currdir + "redirect_io.out", currdir + "redirect_io1.txt")[0])
        self.assertFalse(
            pyutilib.misc.comparison.compare_file(
                currdir + "redirect_io.out",
                currdir + "redirect_io2.txt",
                filter=filter1)[0])
        self.assertFalse(
            pyutilib.misc.comparison.compare_file(
                currdir + "redirect_io.out",
                currdir + "redirect_io2.txt",
                filter=filter2)[0])
        os.remove(currdir + "redirect_io.out")

    def test_redirect2(self):
        # Verify that IO redirection will create an empty file is no output is generated
        pyutilib.misc.setup_redirect(currdir + "redirect_io.out")
        pyutilib.misc.reset_redirect()
        if not os.path.exists(currdir + "redirect_io.out"):
            self.fail("Redirection did not create an empty file.")
        self.assertFalse(
            pyutilib.misc.comparison.compare_file(
                currdir + "redirect_io.out", currdir + "redirect_io2.txt")[0])
        os.remove(currdir + "redirect_io.out")

    def test_redirect3(self):
        # Verify that IO redirection can be nested
        pyutilib.misc.setup_redirect(currdir + "redirect_io1.out")
        print("HERE")
        pyutilib.misc.setup_redirect(currdir + "redirect_io3.out")
        print("THERE")
        print([4, 5, 6])
        pyutilib.misc.reset_redirect()
        print([1, 2, 3])
        #
        # Force a flush to ensure code coverage
        #
        sys.stdout.flush()
        pyutilib.misc.reset_redirect()
        self.assertFalse(
            pyutilib.misc.comparison.compare_file(
                currdir + "redirect_io1.out", currdir + "redirect_io1.txt")[0])
        os.remove(currdir + "redirect_io1.out")
        self.assertFalse(
            pyutilib.misc.comparison.compare_file(
                currdir + "redirect_io3.out", currdir + "redirect_io3.txt")[0])
        self.assertFalse(
            pyutilib.misc.comparison.compare_file(
                currdir + "redirect_io3.out",
                currdir + "redirect_io3.txt",
                filter=filter3)[0])
        os.remove(currdir + "redirect_io3.out")

    def test_redirect4(self):
        # Verify that IO redirection works with file-like objects
        output = six.StringIO()
        pyutilib.misc.setup_redirect(output)
        print("HERE")
        print([1, 2, 3])
        #
        # Force a flush to ensure code coverage
        #
        sys.stdout.flush()
        pyutilib.misc.reset_redirect()
        self.assertEqual(output.getvalue(), "HERE\n[1, 2, 3]\n")

    def test_format_io(self):
        # Test that formated IO looks correct.
        pyutilib.misc.setup_redirect(currdir + "format_io.out")
        print(pyutilib.misc.format_io(0.0))
        print(pyutilib.misc.format_io(0))
        print(pyutilib.misc.format_io(1e-1))
        print(pyutilib.misc.format_io(1e+1))
        print(pyutilib.misc.format_io(1e-9))
        print(pyutilib.misc.format_io(1e+9))
        print(pyutilib.misc.format_io(1e-99))
        print(pyutilib.misc.format_io(1e+99))
        print(pyutilib.misc.format_io(1e-100))
        print(pyutilib.misc.format_io(1e+100))
        print(pyutilib.misc.format_io(-1e-1))
        print(pyutilib.misc.format_io(-1e+1))
        print(pyutilib.misc.format_io(-1e-9))
        print(pyutilib.misc.format_io(-1e+9))
        print(pyutilib.misc.format_io(-1e-99))
        print(pyutilib.misc.format_io(-1e+99))
        print(pyutilib.misc.format_io(-1e-100))
        print(pyutilib.misc.format_io(-1e+100))
        print(pyutilib.misc.format_io('string'))
        pyutilib.misc.reset_redirect()
        self.assertFalse(
            pyutilib.misc.comparison.compare_file(currdir + "format_io.out",
                                                  currdir + "format_io.txt")[0])
        os.remove(currdir + "format_io.out")

    def test_format_float_err1(self):
        # Test that errors are generated for non floats
        try:
            pyutilib.misc.format_float('1')
            self.fail("Should have thrown a TypeError exception")
        except TypeError:
            pass

    def test_indenter_write(self):
        output = six.StringIO()
        indenter = pyutilib.misc.StreamIndenter(output, "X")
        indenter.write("foo")
        self.assertEqual(output.getvalue(), "Xfoo")
        indenter.write("\n")
        self.assertEqual(output.getvalue(), "Xfoo\n")
        indenter.write("foo\n")
        self.assertEqual(output.getvalue(), "Xfoo\nXfoo\n")
        indenter.write("")
        self.assertEqual(output.getvalue(), "Xfoo\nXfoo\n")
        indenter.write("baz\nbin")
        self.assertEqual(output.getvalue(), "Xfoo\nXfoo\nXbaz\nXbin")

        self.assertEqual(indenter.closed, False)
        indenter.close()
        self.assertEqual(indenter.closed, True)

    def test_indenter_writelines(self):
        output = six.StringIO()
        indenter = pyutilib.misc.StreamIndenter(output, "X")
        indenter.writelines(["foo"])
        self.assertEqual(output.getvalue(), "Xfoo")
        indenter.writelines(["foo\n", "bar\n"])
        self.assertEqual(output.getvalue(), "Xfoofoo\nXbar\n")
        indenter.writelines(["", ""])
        self.assertEqual(output.getvalue(), "Xfoofoo\nXbar\n")

        self.assertEqual(indenter.closed, False)
        indenter.close()
        self.assertEqual(indenter.closed, True)

    def test_redirected_stdout(self):
        with pyutilib.misc.capture_output() as output:
            print("test1")
            print("test2")
        self.assertEqual(output.getvalue(), "test1\ntest2\n")

    def test_redirected_stdout_with_arg(self):
        stream = six.StringIO()
        stream.write("before\n")
        with pyutilib.misc.capture_output(stream) as output:
            print("test1")
            print("test2")
        self.assertEqual(output.getvalue(), "before\ntest1\ntest2\n")


if __name__ == "__main__":
    unittest.main()



#
# Unit Tests for pyutilib.options.configuration
#
#

import glob
import sys
import shutil
import os
from os.path import abspath, dirname
currdir = dirname(abspath(__file__)) + os.sep
tempdir = dirname(abspath(__file__)) + os.sep + 'tempdir' + os.sep

from pyutilib.component.config import *
import pyutilib.th as unittest

old_tempdir = TempfileManager.tempdir


class Test(unittest.TestCase):

    def setUp(self):
        PluginGlobals.add_env("testing.options")
        TempfileManager.tempdir = tempdir
        TempfileManager.push()
        if os.path.exists(tempdir):
            shutil.rmtree(tempdir)
        os.mkdir(tempdir)

    def tearDown(self):
        TempfileManager.pop()
        TempfileManager.tempdir = old_tempdir
        PluginGlobals.remove_env("testing.options")
        if os.path.exists(tempdir):
            shutil.rmtree(tempdir)

    def test_add1(self):
        """Test explicit adding of a file that is missing"""
        try:
            TempfileManager.add_tempfile(tempdir + 'add1')
            self.fail("Expected IOError because file 'add1' does not exist")
        except IOError:
            pass

    def test_add1_dir(self):
        """Test explicit adding of a directory that is missing"""
        try:
            TempfileManager.add_tempfile(tempdir + 'add1')
            self.fail(
                "Expected IOError because directory 'add1' does not exist")
        except IOError:
            pass

    def test_add2(self):
        """Test explicit adding of a file that is missing"""
        TempfileManager.add_tempfile(tempdir + 'add2', False)

    def test_add2_dir(self):
        """Test explicit adding of a directory that is missing"""
        TempfileManager.add_tempfile(tempdir + 'add2', False)

    def test_add3(self):
        """Test explicit adding of a file that already exists"""
        OUTPUT = open(tempdir + 'add3', 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        TempfileManager.add_tempfile(tempdir + 'add3')

    def test_add3_dir(self):
        """Test explicit adding of a directory that already exists"""
        os.mkdir(tempdir + 'add3')
        TempfileManager.add_tempfile(tempdir + 'add3')

    def test_pushpop1(self):
        """Test pushpop logic"""
        TempfileManager.push()
        OUTPUT = open(tempdir + 'pushpop1', 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        TempfileManager.add_tempfile(tempdir + 'pushpop1')
        TempfileManager.pop()
        if os.path.exists(tempdir + 'pushpop1'):
            self.fail("pop() failed to clean out files")

    def test_pushpop1_dir(self):
        """Test pushpop logic with directories"""
        TempfileManager.push()
        os.mkdir(tempdir + 'pushpop1')
        TempfileManager.add_tempfile(tempdir + 'pushpop1')
        TempfileManager.pop()
        if os.path.exists(tempdir + 'pushpop1'):
            self.fail("pop() failed to clean out directories")

    def test_pushpop2(self):
        """Test pushpop logic"""
        TempfileManager.push()
        OUTPUT = open(tempdir + 'pushpop2', 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        TempfileManager.add_tempfile(tempdir + 'pushpop2')

        TempfileManager.push()
        OUTPUT = open(tempdir + 'pushpop2a', 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        TempfileManager.add_tempfile(tempdir + 'pushpop2a')
        TempfileManager.pop()
        if not os.path.exists(tempdir + 'pushpop2'):
            self.fail("pop() clean out all files")
        if os.path.exists(tempdir + 'pushpop2a'):
            self.fail("pop() failed to clean out files")

        TempfileManager.pop()
        if os.path.exists(tempdir + 'pushpop2'):
            self.fail("pop() failed to clean out files")

    def test_pushpop2_dir(self):
        """Test pushpop logic with directories"""
        TempfileManager.push()
        os.mkdir(tempdir + 'pushpop2')
        TempfileManager.add_tempfile(tempdir + 'pushpop2')

        TempfileManager.push()
        os.mkdir(tempdir + 'pushpop2a')
        TempfileManager.add_tempfile(tempdir + 'pushpop2a')
        TempfileManager.pop()
        if not os.path.exists(tempdir + 'pushpop2'):
            self.fail("pop() clean out all files")
        if os.path.exists(tempdir + 'pushpop2a'):
            self.fail("pop() failed to clean out files")

        TempfileManager.pop()
        if os.path.exists(tempdir + 'pushpop2'):
            self.fail("pop() failed to clean out files")

    def test_clear(self):
        """Test clear logic"""
        TempfileManager.push()
        OUTPUT = open(tempdir + 'pushpop2', 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        os.mkdir(tempdir + 'pushpopdir2')
        TempfileManager.add_tempfile(tempdir + 'pushpop2')
        TempfileManager.add_tempfile(tempdir + 'pushpopdir2')

        TempfileManager.push()
        OUTPUT = open(tempdir + 'pushpop2a', 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        os.mkdir(tempdir + 'pushpopdir2a')
        TempfileManager.add_tempfile(tempdir + 'pushpop2a')
        TempfileManager.add_tempfile(tempdir + 'pushpopdir2a')

        TempfileManager.clear_tempfiles()

        if os.path.exists(tempdir + 'pushpop2a'):
            self.fail("clear_tempfiles() failed to clean out files")
        if os.path.exists(tempdir + 'pushpopdir2a'):
            self.fail("clear_tempfiles() failed to clean out directories")
        if os.path.exists(tempdir + 'pushpop2'):
            self.fail("clear_tempfiles() failed to clean out files")
        if os.path.exists(tempdir + 'pushpopdir2'):
            self.fail("clear_tempfiles() failed to clean out directories")

    def test_create1(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempfile()
        OUTPUT = open(fname, 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.startswith('tmp'))

    def test_create1_dir(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempdir()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.startswith('tmp'))

    def test_create1a(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempfile(dir=tempdir)
        OUTPUT = open(fname, 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.startswith('tmp'))

    def test_create1a_dir(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempdir(dir=tempdir)
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.startswith('tmp'))

    def test_create2(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempfile(prefix='foo')
        OUTPUT = open(fname, 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.startswith('foo'))

    def test_create2_dir(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempdir(prefix='foo')
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.startswith('foo'))

    def test_create3(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempfile(suffix='bar')
        OUTPUT = open(fname, 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.endswith('bar'))

    def test_create3_dir(self):
        """Test create logic - no options"""
        fname = TempfileManager.create_tempdir(suffix='bar')
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertTrue(fname.endswith('bar'))

    def test_create4(self):
        """Test create logic - no options"""
        TempfileManager.sequential_files(2)
        fname = TempfileManager.create_tempfile()
        OUTPUT = open(fname, 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertEqual(fname, 'tmp2')
        #
        TempfileManager.unique_files()
        fname = TempfileManager.create_tempfile()
        OUTPUT = open(fname, 'w')
        OUTPUT.write('tempfile\n')
        OUTPUT.close()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 2)
        fname = os.path.basename(fname)
        self.assertNotEqual(fname, 'tmp3')
        self.assertTrue(fname.startswith('tmp'))

    def test_create4_dir(self):
        """Test create logic - no options"""
        TempfileManager.sequential_files(2)
        fname = TempfileManager.create_tempdir()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 1)
        fname = os.path.basename(fname)
        self.assertEqual(fname, 'tmp2')
        #
        TempfileManager.unique_files()
        fname = TempfileManager.create_tempdir()
        self.assertEqual(len(list(glob.glob(tempdir + '*'))), 2)
        fname = os.path.basename(fname)
        self.assertNotEqual(fname, 'tmp3')
        self.assertTrue(fname.startswith('tmp'))


if __name__ == "__main__":
    unittest.main()



import os
import sys
import pyutilib.component.app

currdir = sys.argv[-1] + os.sep

app = pyutilib.component.app.SimpleApplication("foo")
app.config.load(currdir + "log1.ini")
#print type(app.logger.log_dir)
app.logger.log_dir = currdir
#print type(app.logger.log_dir)
app.config.save(currdir + "tmp2.ini")



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
"""
The Option class provides a mechanism for application classes to transparently
load configuration data.  For example, the consider the following class declaration:

  class FOO(Option):

    declare_option("x", local_name="bar")

    def __init__(self):
        self.x = FOO.bar

The FOO class contains a declaration 'bar', which is initialized by loading
a Configuration object.

"""

import weakref
import sys
import re
import os
import os.path
from pyutilib.component.core import PluginError, Interface, implements, Plugin, PluginGlobals, ExtensionPoint, IOptionDataProvider
try:
    unicode
except:
    basestring = str


class OptionError(PluginError):
    """Exception raised when there is an error processing an option."""


class IOption(Interface):
    """An interface that supports the initialization of Option objects
    from a Configuration object."""

    def matches_section(self, section):
        """Returns True if this option matches the section name."""

    def matches_name(self, section):
        """Returns True if this option matches the option name."""

    def load(self, option, value):
        """Loads this option, if it matches the option name.  Returns True
        if the option is loaded, and False otherwise."""

    def get_value(self):
        """Return the value of this option"""

    def default_str(self):
        """Return a string value that describes the default option value"""

    def reset(self):
        """Reset all options to their default values."""


class IUpdatedOptionsAction(Interface):
    """Define actions that are applied after options have been updated."""

    def reset_after_updates(self):
        """Perform actions that use updated option data."""


class IFileOption(Interface):
    """An interface that supports the initialization of the directory for
    options that specify files.  This is needed to correctly initialize
    relative paths for files."""

    def set_dir(self, path):
        """Sets the directory of the configuration data file."""


class OptionData(Plugin):
    """
    A class that is used to share option data between Option objects.
    This data in this class represents option data as

        section -> option -> data

    Note: this class does not currentl support the situation where an
    option is removed.  This would require a backwards map from option
    names to the Option classes that declare them.
    """

    implements(IOptionDataProvider, service=True)

    def __init__(self):
        """Constructor"""
        self.ignore_missing = False
        self.data = {}

    def get_data(self):
        """Get the class data"""
        return self.data

    def set(self, section, name, value):
        """Set the value of an option in a specified section"""
        if section not in self.data:
            self.data[section] = {}
        self.data[section][name] = value

    def get(self, section, name):
        """Get the value of an option in a specified section"""
        try:
            return self.data[section][name]
        except Exception:
            if self.ignore_missing:
                return None
            if section in self.data:
                raise OptionError(
                    "Problem retrieving the value of option %r from section %r. Valid keys are %s"
                    % (name, section, self.data[section].keys()))
            else:
                raise OptionError(
                    "Problem retrieving the value of option %r from section %r. Undefined section."
                    % (name, section))

    def clear(self, keys=None):
        """Clears the data"""
        if keys is None:
            self.data = {}
        else:
            for key in keys:
                if not key in self.data:
                    continue
                del self.data[key]


class OptionPlugin(Plugin):
    """Manages the initialization of an Option."""

    implements(IOption, service=True)

    def __init__(self):
        """
        Declare an extension point for a data provider, and
        construct one if one hasn't already been provided.
        """
        self.data = ExtensionPoint(IOptionDataProvider)
        if PluginGlobals._default_OptionData is None:
            PluginGlobals._default_OptionData = OptionData()
        #
        # This is a hack.  We shouldn't need to test if len(self.data) is zero.
        # Somewhere in our tests, the weakref to the OptionData object is being 
        # corrupted.  Perhaps this is caused by 'nose' or 'import' logic?
        #
        if True and len(self.data) == 0:
            PluginGlobals.interface_services[IOptionDataProvider].add(
                PluginGlobals._default_OptionData._id)
            PluginGlobals.plugin_instances[
                PluginGlobals._default_OptionData._id] = weakref.ref(
                    PluginGlobals._default_OptionData)
        #
        if len(self.data) == 0:
            #if False:
            #print "ZZZ", ep.Xextensions()
            #print "HERE", PluginGlobals._default_OptionData._id, PluginGlobals._default_OptionData.ctr
            #print "HERE", PluginGlobals._default_OptionData
            #print "HERE - id", id(PluginGlobals._default_OptionData)
            #print "HERE", getattr(PluginGlobals._default_OptionData, '_HERE_', None)
            #print "HERE", PluginGlobals._default_OptionData.__interfaces__
            #print ""
            #print "HERE", PluginGlobals.interface_services
            #print "HERE", PluginGlobals.plugin_instances.keys()
            #for exe_ in PluginGlobals._executables:
            #print exe_._id, exe_
            #print "LEN", len(PluginGlobals.env)
            #for name_ in PluginGlobals.env:
            #env_ = PluginGlobals.env[name_]
            #print env_.name
            #print env_.nonsingleton_plugins
            #print [env_.singleton_services[cls_] for cls_ in env_.singleton_services]
            raise PluginError(
                "Problem constructing a global OptionData object %s" %
                self.name)

    def matches_section(self, section):
        """
        This method returns true if the section name matches the option
        section, or if the option's section regular expression matches the
        section name.
        """
        return (section == self.section) or (self.section_re != None and (
            not self.section_p.match(section) is None))

    def matches_name(self, name):
        """
        This method returns true if the name matches the options' name.
        """
        return (self.name == "") or (name == self.name)

    def convert(self, value, default):
        """Convert a value into a specific type.  The default behavior is to
        take the list of values, and simply return the last one defined in the
        configuration."""
        return value[-1]

    def get_value(self):
        """
        Get the option value.
        """
        return self.data.service().get(self.section, self.name)

    def set_value(self, _value_, raw=False):
        """
        Set the option value.  By default, the option is converted using
        the option-specific `convert` method, but the `raw` option can be
        specified to force the raw value to be inserted.
        """
        if raw:
            self.data.service().set(self.section, self.name, _value_)
        else:
            if not type(_value_) is list or len(_value_) == 0:
                _value_ = [_value_]
            self.data.service().set(self.section, self.name, self.convert(
                _value_, self.default))

    def load(self, _option_, _value_):
        """
        Load an option value.  This method assumes that the option value is
        provided in a list, which is the format used when interfacing with
        the Configure class.
        """
        if type(_value_) is list and len(_value_) == 0:
            raise OptionError("Attempting to load option %r with empty data" %
                              (self.name))
        try:
            self.set_value(_value_)
        except OptionError:
            err = sys.exc_info()[1]
            raise OptionError("Error loading option %r: %s" %
                              (str(_option_), str(err)))
        return True

    def reset(self):
        """Set option to its default value"""
        self.set_value(self.default, raw=True)

    def default_str(self):
        """Return a string value that describes the default option value"""
        return str(self.default)


class Option(OptionPlugin):
    """Descriptor for configuration options.

    This class uses an OptionPlugin instance to support initialization of
    the option.  However, the OptionPlugin instance is given a handle on
    the Option instance to perform this initialization.  This allows for the
    explicit referencing of this object's data, as well as referencing the
    conversion routines that are specified in subclasses of Option.
    """

    def __init__(self, name=None, **kwds):
        """Create the extension point.

        @param name: the name of the option
        @param section: the name of the configuration section this option
            belongs to
        @param default: the default value for the option
        @param doc: documentation of the option
        """
        self.section = "globals"
        super(Option, self).__init__()
        self.section_re = None
        self.name = name
        self.default = None
        self.__doc__ = ""
        for (k, v) in kwds.items():
            self._parse_option(k, v)
        if name is None:
            raise OptionError("The Option class requires a name")
        self.reset()

    def _parse_option(self, k, v):
        """Parse option"""
        if k == "section":
            self.section = v
        elif k == "section_re":
            self.section_re = v
            self.section_p = re.compile(v)
        elif k == "default":
            self.default = v
        elif k == "doc":
            self.__doc__ = v
        elif not k in ("local_name"):
            #
            # It's convenient to ignore several keywords, which are really used by
            # declare_option().
            #
            raise OptionError("Unknown keyword: " + k)

    def __get__(self, instance, owner):
        raise PluginError(
            "The Option class cannot be used as a class decorator")

    def __set__(self, instance, value):
        raise PluginError(
            "The Option class cannot be used as a class decorator")

    def __repr__(self, simple=False):
        """Returns a string representation of the option name"""
        sec = " [" + self.section + "]"
        return '<%s%s %r>' % (self.__class__.__name__, sec, self.name)


class VirtualOption(object):
    """
    This class is the decorator that is used to declare an option.  To use
    an option, each instance must declare an Option object.
    """

    def __init__(self, name, option=None):
        self.name = name
        self.option = option

    def _get_option(self, instance):
        if not self.option is None:
            return self.option
        try:
            return getattr(instance, self.name)
        except AttributeError:
            raise PluginError(
                "Expected attribute %r in instance %r.  The declare_option() method was not called in the constructor!"
                % (self.name, instance))

    def __get__(self, instance, owner):
        """Returns the value of the option, accessed through the local instance."""
        if owner is None:  #pragma:nocover
            return self
        return self._get_option(instance).get_value()

    def __set__(self, instance, value):
        """Sets the value of the option"""
        self._get_option(instance).set_value(value)

    def __repr__(self, simple=False):
        """Returns a string representation of the option name"""
        return "<VirtualOption %s>" % self.name


def declare_option(name, cls=Option, **kwds):
    if not issubclass(cls, Option):
        raise PluginError(
            "The 'cls' argument must specify an Option class type: %s" % cls)
    local_name = kwds.get("local_name", name)

    frame = sys._getframe(1)
    locals_ = frame.f_locals
    #print "HERE", locals_ is not frame.f_globals, '__module__' in locals_
    if locals_ is not frame.f_globals:
        kwds["name"] = name
        option_ = cls(**kwds)
        if '__module__' in locals_:
            #
            # Class decorator.  Initialize the VirtualOption, and then
            # create a locally-available Option instance.
            #
            locals_[local_name] = VirtualOption("_" + local_name, option_)
        else:
            #
            # Class constructor.  If a VirtualOption does not exist in this
            # class, create it.  Create an Option instance in this instance.
            #
            if not local_name in locals_["self"].__class__.__dict__:
                setattr(locals_["self"].__class__, local_name,
                        VirtualOption("_" + local_name))
            setattr(locals_["self"], "_" + local_name, option_)
    else:
        raise PluginError(
            "declare_option() can only be used in a class definition")


class DictOption(Option):
    """
    An option that supports an interface to _all_ options in a section.
    For example, consider the following class:

        class FOO(object):

            options = DictOption(section="bar")

    The `options` object will be populated by all data in the `bar` section.
    For example, if the values `a` and `b` are in this section, then they can
    be referenced as `foo.a` and `foo.b`.  Similarly, data can be inserted
    into the `bar` section by specifying the value of attributes of options:

        options.c = 1

    """

    class SectionWrapper(object):
        """
        A class that is used to manage attribute access to a DictOption
        instance.  This class uses IOptionDataProvider extension points
        to directly access shared Option data.
        """

        def __init__(self, section, ignore_missing=False):
            self._section_ = section
            ep = ExtensionPoint(IOptionDataProvider)
            ep.service().ignore_missing = ignore_missing
            self.__dict__["data"] = ep

        def __iter__(self):
            if not self._section_ in self.data.service().data:
                return {}.__iter__()
            return self.data.service().data[self._section_].__iter__()

        def __getitem__(self, name):
            return self.data.service().get(self._section_, name)

        def keys(self):
            if not self._section_ in self.data.service().data:
                return []
            return self.data.service().data[self._section_].keys()

        def __getattr__(self, name):
            try:
                return self.__dict__[name]
            except:
                return self.data.service().get(self._section_, name)

        def __setattr__(self, name, value):
            if name[0] == "_":
                self.__dict__[name] = value
            else:
                self.data.service().set(self._section_, name, value)

    def __init__(self, **kwds):
        """Constructor."""
        kwds["name"] = ""
        if 'ignore_missing' in kwds:
            ignore_missing = kwds['ignore_missing']
            del kwds['ignore_missing']
        else:
            ignore_missing = False
        super(DictOption, self).__init__(**kwds)
        self.default = DictOption.SectionWrapper(
            self.section, ignore_missing=ignore_missing)

    def get_value(self):
        """Returns the default value."""
        return self.default

    def set_value(self, _value_, raw=False):
        """
        If the value is a dictionary, then it is evaluated to populate
        corresponding data in the options data.

        NOTE: the raw option is ignored, but maintained for compatibility with the
        parent class.
        """
        if not type(_value_) is dict:
            return
        for key in _value_:
            self.default.__setattr__(key, _value_[key])

    def load(self, _option_, _value_):
        """
        Data is loaded one option/value pair at a time, so this
        method circumvents the core logic of this class.
        """
        self.data.service().set(self.section, _option_, _value_[-1])
        return True

    def default_str(self):
        """Return a string value that describes the default option value"""
        return "None (this dictionary consists of all values in this section)"


class BoolOption(Option):
    """A class that converts option data into boolean values."""

    _TRUE_VALUES = ('yes', 'true', 'enabled', 'on', 'aye', '1', 1, True)

    def convert(self, value, default):
        """Conversion routine."""
        val = value[-1]
        if isinstance(val, basestring):
            val = val.lower() in BoolOption._TRUE_VALUES
        return bool(val)


class IntOption(Option):
    """A class that converts option data into integer values."""

    def convert(self, value, default):
        """Conversion routine."""
        val = value[-1]
        if not val:
            return 0
        try:
            return int(val)
        except ValueError:
            raise OptionError('Expected integer, got %s' % repr(value))
        except TypeError:
            raise OptionError('Expected string or integer type, got %s' %
                              repr(value))


class FloatOption(Option):
    """A class that converts option data into float values."""

    def convert(self, value, default):
        """Conversion routine."""
        val = value[-1]
        if not val:
            return 0
        try:
            return float(val)
        except ValueError:
            raise OptionError('Expected float, got %s' % repr(value))
        except TypeError:
            raise OptionError('Expected string or float type, got %s' %
                              repr(value))


class FileOption(Option):
    """A class that converts option data into a path.  Relative paths are
    converted using the path for the configuration file."""

    implements(IFileOption, service=True)
    implements(IUpdatedOptionsAction, service=True)

    def __init__(self, name, **kwds):
        """
        Constructor.  By default, the current working directory is the
        path used in this data.
        """
        self.dir = None
        Option.__init__(self, name, **kwds)

    def _parse_option(self, k, v):
        """Parse options that are specific to the FileOption class"""
        if k == "directory":
            self.dir = v
        else:
            Option._parse_option(self, k, v)

    def convert(self, value, default):
        """Conversion routine."""
        val = value[-1]
        if not val:
            if default is None:
                return None
            return os.path.normcase(os.path.realpath(default))
        if not os.path.isabs(val):
            val = os.path.join(self.dir, val)
            return os.path.normcase(os.path.realpath(val))
        else:
            return os.path.normcase(os.path.realpath(val))

    def set_dir(self, path):
        """Sets the path of the configuration data file."""
        self.dir = path

    def reset_after_updates(self):
        if self.dir is None:  #pragma:nocover
            raise OptionError("FileOption must have a directory specified.")


class ExecutableOption(FileOption):

    def convert(self, value, default):
        """Conversion routine."""
        val = FileOption.convert(self, value, default)
        if val is None or not os.access(val, os.X_OK):
            raise OptionError(
                "ExecutableOption value %r is not a file that can be executed" %
                val)
        return val

    def convert(self, value, default):
        """Conversion routine."""
        val = value[-1]
        if not val:
            if default is None:
                return None
            val = os.path.normcase(os.path.realpath(default))
        elif os.path.isabs(val):
            val = os.path.normcase(os.path.realpath(val))
        elif not self.dir is None:
            val = os.path.normcase(
                os.path.realpath(os.path.join(self.dir, val)))
        else:
            flag = False
            for path in os.environ.get("PATH", []).split(os.pathsep):
                if os.path.exists(os.path.join(path, val)):
                    val = os.path.normcase(
                        os.path.realpath(os.path.join(path, val)))
                    flag = True
                    break
            if not flag:
                return None
        #
        # Confirm that the executable can be solved
        #
        if val is None or not os.access(val, os.X_OK):
            raise OptionError(
                "ExecutableOption value %r is not a file that can be executed" %
                val)

    def reset_after_updates(self):
        pass



#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
#

from pyutilib.math.numtypes import infinity, nan, is_nan, is_finite
from pyutilib.math.util import approx_equal, as_number, isint, argmax, argmin, mean, median, factorial, perm, isclose



import os
import sys
import glob
import optparse

import pyutilib.subprocess
from pyutilib.th import TestCase

from os.path import dirname

if sys.platform.startswith('win'):
    platform = 'win'
    use_exec = False
else:
    platform = 'linux'
    use_exec = True


def run(package, basedir, argv, use_exec=use_exec, env=None):
    if type(package) not in (list, tuple):
        package = [package]

    parser = optparse.OptionParser(usage='run [OPTIONS] <dirs>')

    parser.add_option(
        '-v',
        '--verbose',
        action='store_true',
        dest='verbose',
        default=False,
        help='Verbose output')
    parser.add_option(
        '--cat',
        '--category',
        action='append',
        dest='cat',
        default=[],
        help='Specify the test category.')
    parser.add_option(
        '--cov',
        '--coverage',
        action='store_true',
        dest='coverage',
        default=False,
        help='Enable the computation of coverage information')
    parser.add_option(
        '--cover-erase',
        action='store_true',
        dest='cover_erase',
        default=False,
        help='Erase any previous coverage data files')
    parser.add_option(
        '-d',
        '--dir',
        action='store',
        dest='dir',
        default=None,
        help='Top-level source directory where the tests are applied.')
    parser.add_option(
        '-p',
        '--package',
        action='store',
        dest='pkg',
        default=package[0],
        help='Limit the coverage to this package')
    parser.add_option(
        '-o',
        '--output',
        action='store',
        dest='output',
        default=None,
        help='Redirect output to a file')
    parser.add_option('--with-doctest',
        action='store_true',
        dest='doctests',
        default=False,
        help='Run tests included in Sphinx documentation')
    parser.add_option('--doc-dir',
        action='store',
        dest='docdir',
        default=None,
        help='Top-level source directory for Sphinx documentation')
    parser.add_option('--no-xunit',
        action='store_false',
        dest='xunit',
        default=True,
        help='Disable the nose XUnit plugin')
    parser.add_option('--dry-run',
        action='store_true',
        dest='dryrun',
        default=False,
        help='Dry run: collect but do not execute the tests')

    options, args = parser.parse_args(argv)

    if env is None:
        env = os.environ.copy()

    if options.output:
        options.output = os.path.abspath(options.output)

    if options.dir is None:
        os.chdir(basedir)
    else:
        os.chdir(options.dir)

    CWD = os.getcwd()
    print("Running tests in directory %s" % (CWD,))

    if platform == 'win':
        binDir = os.path.join(sys.exec_prefix, 'Scripts')
        nosetests = os.path.join(binDir, 'nosetests.exe')
        #
        # JDS [2 Oct 2017]: I am not sure why this was here.  If we find
        # we need it, we can re-add it with an explanation as to why
        # Windows needs special PYTHONPATH handling.
        #
        #srcdirs = []
        #for dir in glob.glob('*'):
        #    if os.path.isdir(dir):
        #        srcdirs.append(os.path.abspath(dir))
        #if 'PYTHONPATH' in env:
        #    srcdirs.append(env['PYTHONPATH'])
        #env['PYTHONPATH'] = os.pathsep.join(srcdirs)
    else:
        binDir = os.path.join(sys.exec_prefix, 'bin')
        nosetests = os.path.join(binDir, 'nosetests')

    if os.path.exists(nosetests):
        cmd = [nosetests]
    else:
        cmd = ['nosetests']

    if binDir not in env['PATH']:
        env['PATH'] = os.pathsep.join([binDir, env.get('PATH','')])

    if options.coverage:
        cmd.append('--with-coverage')
        if options.cover_erase:
            cmd.append('--cover-erase')
        if options.pkg:
            cmd.append('--cover-package=%s' % options.pkg)
        #env['COVERAGE_FILE'] = os.path.join(CWD, '.coverage')

    if options.verbose:
        cmd.append('-v')
    if options.dryrun:
        cmd.append('--collect-only')

    if options.doctests:
        cmd.extend(['--with-doctest', '--doctest-extension=.rst'])
        if options.docdir:
            docdir = os.path.abspath(_options.docdir)
            if not os.path.exists(docdir):
                raise ValueError("Invalid documentation directory, "
                                 "path does not exist")

    if options.xunit:
        cmd.append('--with-xunit')
        cmd.append('--xunit-file=TEST-' + package[0] + '.xml')

    attr = []
    _with_performance = False
    if 'PYUTILIB_UNITTEST_CATEGORY' in env:
        _categories = TestCase.parse_categories(
            env.get('PYUTILIB_UNITTEST_CATEGORY', '') )
    else:
        _categories = []
        for x in options.cat:
            _categories.extend( TestCase.parse_categories(x) )

    # If no one specified a category, default to "smoke" (and anything
    # not built on pyutilib.th.TestCase)
    if not _categories:
        _categories = [ (('smoke',1),), (('pyutilib_th',0),) ]
    # process each category set (that is, each conjunction of categories)
    for _category_set in _categories:
        _attrs = []
        # "ALL" deletes the categories, and just runs everything.  Note
        # that "ALL" disables performance testing
        if ('all', 1) in _category_set:
            _categories = []
            _with_performance = False
            attr = []
            break
        # For each category set, unless the user explicitly says
        # something about fragile, assume that fragile should be
        # EXCLUDED.
        if ('fragile',1) not in _category_set \
           and ('fragile',0) not in _category_set:
            _category_set = _category_set + (('fragile',0),)
        # Process each category in the conjection and add to the nose
        # "attrib" plugin arguments
        for _category, _value in _category_set:
            if not _category:
                continue
            if _value:
                _attrs.append(_category)
            else:
                _attrs.append("(not %s)" % (_category,))
            if _category == 'performance' and _value == 1:
                _with_performance = True
        if _attrs:
            #attr.append('--eval-attr')
            attr.append("--eval-attr=%s" % (' and '.join(_attrs),))
    cmd.extend(attr)
    if attr:
        print(" ... for test categor%s: %s" %
              ('y' if len(attr)<=2 else 'ies',
               ' '.join(attr[1::2])))

    if _with_performance:
        cmd.append('--with-testdata')
        env['NOSE_WITH_TESTDATA'] = '1'
        env['NOSE_WITH_FORCED_GC'] = '1'

    targets = set()
    if len(args) <= 1:
        targets.update(package)
    else:
        for arg in args[1:]:
            if '*' in arg or '?' in arg:
                targets.update(glob.glob(arg))
            else:
                targets.add(arg)
    cmd.extend(list(targets))

    print("Running...\n    %s\n" % (
            ' '.join( (x if ' ' not in x else '"'+x+'"') for x in cmd ), ))
    rc = 0
    if sys.platform.startswith('java'):
        import subprocess
        p = subprocess.Popen(cmd, env=env)
        p.wait()
        rc = p.returncode
    elif options.output:
        sys.stdout.write("Redirecting output to file '%s' ..." % options.output)
        rc, _ = pyutilib.subprocess.run(cmd, env=env, outfile=options.output)
    elif use_exec and not (
            sys.platform.startswith('win')
            and sys.version_info[:2] in ((3,4),(3,5)) ):
        # NOTE: execvpe seems to generate a fatal error on Windows with
        # 3.4 and 3.5.
        #
        # In other Python versions it fails to return the new process'
        # return code to the owning shell (so the build harness doesn't
        # see the command "fail" when there are failing tests.
        rc = None
        sys.stderr.flush()
        sys.stdout.flush()
        os.execvpe(cmd[0], cmd, env)
    else:
        sys.stdout.flush()
        rc, _ = pyutilib.subprocess.run(cmd, env=env, ostream=sys.stdout)
    return rc


def runPyUtilibTests(argv=None, use_exec=use_exec):
    if argv is None:
        argv = sys.argv

    return pyutilib.dev.runtests.run(
        'pyutilib',
        dirname(dirname(dirname(os.path.abspath(__file__)))),
        argv,
        use_exec=use_exec )
